#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 20 13:39:58 2021

@author: mike
"""
import os
import numpy as np
import xarray as xr
import pandas as pd
import tethys_utils as tu
from tethysts.utils import s3_client, get_object_s3, read_json_zstd, b2_public_key_pattern, read_pkl_zstd
import tethys_data_models as tdm
from typing import List, Optional, Dict, Union
# import boto3
# import botocore
from botocore import exceptions as bc_exceptions
import requests
import uuid
from cryptography.fernet import Fernet
import pandas as pd
from pydantic import HttpUrl
import concurrent.futures
import copy
from shapely import wkb, wkt, geometry
import pathlib
import stat
import dask
from hashlib import blake2b
import glob
from datetime import date, datetime
import multiprocessing as mp
import pyproj
import random

##############################################
### Parameters

b2_dl_by_id = 'https://b2.tethys-ts.xyz/b2api/v1/b2_download_file_by_id?fileId={obj_id}'


preprocessed_dir = 'preprocessed_data'
previous_dir = 'previous_results'
final_dir = 'final_results'

preprocessed_file_str = '{ds_id}_{file_id}.nc.zst'
interim_file_str = '{ds_id}_{stn_id}_{date}.nc'

base_encoding = {'lon': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 1e-07},
'lat': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 1e-07},
'altitude': {'dtype': 'int32', '_FillValue': -9999, 'scale_factor': 0.001},
'time': {'_FillValue': -99999999, 'units': 'days since 1970-01-01 00:00:00'},
'height': {'dtype': 'int16', '_FillValue': -9999}}

dask.config.set(**{'array.slicing.split_large_chunks': False})

###############################################
### Helper functions



def save_chunks(chunks_list, output_path):
    """

    """
    for c in chunks_list:
        chunk_json = tdm.dataset.ChunkID(height=int(c.height.values[0] * 1000), start_date=int(c.attrs['start_date'])).json(exclude_none=True).encode('utf-8')
        chunk_id = blake2b(chunk_json, digest_size=12).hexdigest()
        stn_id = str(c.station_id.values[0])
        file_path = os.path.join(output_path, stn_id + '_' + chunk_id + '.nc.zst')

        b = c.copy().load()

        obj1 = tu.misc.write_pkl_zstd(b.to_netcdf(), file_path)

        b.close()
        del b
        del obj1


def multi_save_dataset_stations(nc_paths, block_length, block_length_factor: int = 10, compression='zstd', remove_station_data=True, max_workers=3):
    """

    """
    if isinstance(nc_paths, str):
        nc_paths1 = glob.glob(nc_paths)
    elif isinstance(nc_paths, list):
        nc_paths1 = nc_paths

    ## Iterate through files
    if max_workers <= 1:

        nc_paths = []
        for nc_path in nc_paths1:
            new_paths0 = tu.processing.save_dataset_stations(nc_path, block_length * block_length_factor, compression=compression, remove_station_data=remove_station_data)
            nc_paths.append(new_paths0)
    else:
        with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
            futures = []
            for nc_path in nc_paths1:
                f = executor.submit(tu.processing.save_dataset_stations, nc_path, block_length * block_length_factor, compression=compression, remove_station_data=remove_station_data)
                futures.append(f)
            runs = concurrent.futures.wait(futures)

        new_paths = [r.result() for r in runs[0]]

    ## process output
    new_paths1 = []
    for new_path in new_paths:
        new_paths1.extend(new_path)

    new_paths1.sort()

    return new_paths1


def estimate_time_interval_accurate(data, block_length, null_grid=None, max_mb=2):
    """

    """
    ## Get the dimension data
    dims = dict(data.dims)

    ## Test requested block_length
    chunks_list = tu.processing.chunk_data(data, block_length=block_length, time_interval=None, null_grid=null_grid)

    dim_sizes = np.array([np.prod(list(c.dims.values())) for c in chunks_list])
    max_dim_index = np.argsort(dim_sizes)[-10:]
    sum_dim_objs = np.sum(dim_sizes[max_dim_index])

    sum_chunks = 0
    for i in max_dim_index:
        chunk = chunks_list[i].copy().load()
        obj_len = len(tu.misc.write_pkl_zstd(chunk.to_netcdf()))

        sum_chunks += obj_len

        chunk.close()
        del chunk

    dim_per_mb = int(sum_dim_objs/(sum_chunks * 0.000001))

    if 'geometry' in dims:
        geo_sizes = np.array([c.dims['geometry'] for c in chunks_list])
    else:
        geo_sizes = np.array([c.dims['lon'] * c.dims['lat'] for c in chunks_list])

    max_geo_size = np.max(geo_sizes)

    times_per_mb = (dim_per_mb/max_geo_size)

    ## Calc time_interval
    times1 = pd.to_datetime(data['time'].values)
    days = times1.floor('D').drop_duplicates()
    times_per_day = len(times1)/len(days)

    days_per_mb = times_per_mb/times_per_day

    days_per_chunk = int(days_per_mb * max_mb)

    ## Test final parameters
    chunks_list = tu.processing.chunk_data(data, block_length=block_length, time_interval=days_per_chunk, null_grid=null_grid)

    n_chunks = len(chunks_list)

    _ = [c.close() for c in chunks_list]
    del chunks_list

    output_dict = {'time_interval': days_per_chunk,
                   'n_geometries_per_station': max_geo_size,
                   'n_chunks': n_chunks,
                   'values_per_mb': dim_per_mb
                   }

    return output_dict


def bulk_estimate_time_interval(data, block_lengths, null_grid=None, max_mb=2):
    """

    """
    block_length_dict = {}
    for b in block_lengths:
        print(b)
        o1 = estimate_time_interval_accurate(data, b, null_grid, max_mb)
        block_length_dict[b] = o1

    return block_length_dict


def estimate_time_interval_rough(data_grid_length, time_freq, block_lengths, max_mb=2, values_per_mb=550000):
    """
    Function to roughly estimate the appropriate time intervals based on specific block lengths.

    Parameters
    ----------
    data_grid_length: float or int
        The grid resolution.
    time_freq: str
        The time resolution in pandas time freq format.
    block_lengths: list of float
        The block lengths to test. Should be equal to or greater than the data_grid_length.
    max_mb: float or int
        The max size of the results object.
    values_per_mb: int
        The number of data values in a results object per MB. Do not change unless you've done the testing to determine this value.
    """
    t1 = pd.date_range('2000-01-01', '2000-01-08', freq = time_freq)[:-1]
    val_per_day = len(t1)/7

    total_val_per_day = max_mb * values_per_mb

    res_list = []
    for bl in block_lengths:
        n_stns = int(np.ceil((bl**2)/(data_grid_length**2)))
        days_per_chunk = int(total_val_per_day/n_stns/val_per_day)
        dict1 = {'time_interval': days_per_chunk,
                 'n_geometries_per_station': n_stns}
        res_list.append(dict1)

    res_dict = {bl: res_list[i] for i, bl in enumerate(block_lengths)}

    return res_dict


def decompress_path(glob_path, compression_type='gzip', max_workers=4, **kwargs):
    """

    """
    if isinstance(glob_path, str):
        files1 = glob.glob(glob_path)
    elif isinstance(glob_path, list):
        files1 = glob_path
    new_files = tu.data_io.decompress_files_remove(files1, max_workers=max_workers)

    return new_files


def file_format_conversion(glob_path, file_format='grib', max_workers=4, **kwargs):
    """
    Function to convert data files to netcdf files.
    This also should add in the time coordinate when converting TIF files. It should parse the time from the file name.
    """
    if isinstance(glob_path, str):
        files1 = glob.glob(glob_path)
    elif isinstance(glob_path, list):
        files1 = glob_path
    new_files = tu.data_io.convert_gribs_remove(files1, max_workers=max_workers)

    return new_files


def get_obj_list(glob_path, date_format=None, freq=None, from_date=None, to_date=None, connection_config=None, bucket=None):
    """

    """
    glob_path2 = glob_path

    if isinstance(connection_config, dict) and isinstance(bucket, str):
        if glob_path2.startswith('/'):
            glob_path2 = glob_path2[1:]
        glob_path3, glob_ext = os.path.split(glob_path2)
        if not glob_path3.endswith('/'):
            glob_path3 = glob_path3 + '/'
        s3 = s3_client(connection_config)
        obj_list = tu.s3.list_objects_s3(s3, bucket, glob_path3, delimiter='/', date_format=date_format)
        obj_list1 = obj_list.rename(columns={'Key': 'path', 'Size': 'size', 'KeyDate': 'date'}).drop(['LastModified', 'ETag'], axis=1).copy()
        # obj_list1['remote_type'] = 's3'
    else:
        files1 = glob.glob(glob_path)
        sizes = [os.path.getsize(f) for f in files1]
        obj_list1 = pd.DataFrame(zip(files1, sizes), columns=['path', 'size'])
        # obj_list1['remote_type'] = 'local'

        if isinstance(date_format, str):
            dates1 = tu.misc.filter_data_paths(obj_list1.path, date_format, return_dates=True)
            obj_list1 = pd.merge(obj_list1, dates1, on='path')

    filter1 = [pathlib.Path(p).match(glob_path2) for p in obj_list1.path.values]
    obj_list2 = obj_list1.iloc[filter1]

    if isinstance(date_format, str):
        if isinstance(from_date, (str, pd.Timestamp, date, datetime)):
            obj_list2 = obj_list2[obj_list2['date'] >= pd.Timestamp(from_date)].copy()
        if isinstance(to_date, (str, pd.Timestamp, date, datetime)):
            obj_list2 = obj_list2[obj_list2['date'] < pd.Timestamp(to_date)].copy()

    if isinstance(freq, str):
        grp1 = obj_list2.groupby(pd.Grouper(key='date', freq=freq, origin='start'))

        obj_list3 = [v for k, v in grp1]
    else:
        obj_list3 = obj_list2

    return obj_list3


def copy_source_objs(source_paths, dest_path, bucket=None, connection_config=None, public_url=None, compression=None, threads=None, max_workers=None):
    """

    """
    objs = source_paths

    if isinstance(connection_config, (dict, str)) and isinstance(bucket, str):
        s3 = s3_client(connection_config, threads)

        if isinstance(max_workers, int):

            with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
                futures = []
                for obj in objs:
                    f = executor.submit(tu.data_io.copy_s3_file, source_path=obj, dest_path=dest_path, connection_config=connection_config, bucket=bucket, public_url=public_url, compression=compression)
                    futures.append(f)
                runs = concurrent.futures.wait(futures)

            new_paths = [r.result() for r in runs[0]]

        else:
            if threads is None:
                threads = 4

            with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
                futures = []
                for obj in objs:
                    f = executor.submit(tu.data_io.copy_s3_file, source_path=obj, dest_path=dest_path, s3=s3, connection_config=connection_config, bucket=bucket, public_url=public_url, compression=compression)
                    futures.append(f)
                runs = concurrent.futures.wait(futures)

            new_paths = [r.result() for r in runs[0]]

    else:
        new_paths = []
        for obj in objs:
            path1 = tu.data_io.copy_file(source_path=obj, dest_path=dest_path, compression=compression)
            new_paths.append(path1)

    new_paths.sort()

    return new_paths


def copy_interim_objs(source_paths, dest_path, bucket, connection_config=None, public_url=None, threads=20):
    """

    """
    s3 = s3_client(connection_config, threads)

    path1 = pathlib.Path(dest_path)
    with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
        futures = []
        for source_path in source_paths:
            p_list = source_path.split('/')
            ds_id = p_list[4]
            stn_id, date, nc, z = p_list[5].split('.')

            file_name = interim_file_str.format(ds_id=ds_id, stn_id=stn_id, date=date)
            dest_path1 = str(path1.joinpath(file_name))

            f = executor.submit(tu.data_io.copy_s3_file, source_path, dest_path1, bucket, s3=s3, public_url=public_url, compression='zstd')
            futures.append(f)
        runs = concurrent.futures.wait(futures)

    new_paths = [r.result() for r in runs[0]]

    new_paths.sort()

    return new_paths


def variable_processing(nc_paths, time_name, x_name, y_name, variables, projected_coordinates=True, nc_source='normal', max_workers=4):
    """
    The variable processing function does three things:
        - select only the specific variables that should be exported,
        - restructure (transpose in xarray) the coordinates so that they are ordered by time, y, x, height,
        - remove duplicate overlapping timestamps between files.

    The restructuring of the coordinates must make sure that the time coordinate is called "time", but the x and y coordinates do not necessarily need to be modified. They just need to be in the correct order. If they are actually longitude and latitude, changing them to lon and lat will make things easier in the future.
    This process will iterate over all of the nc files rather than trying to open them and process them using Dask and xarray. They will initially be opened in Dask to determine the overlapping times only.

    Parameters
    ----------
    nc_paths : list of str or glob str
        The paths to the nc files to be processed.
    variables : list of str
        The variables that should be extracted from the nc files.
    time_index_bool : list of bool
        The boolean time index in case only one file is pulled down to be processed. Normally this should not be needed as there will be multiple files to determine the overlapping times.
    """
    if isinstance(nc_paths, str):
        nc_paths1 = glob.glob(nc_paths)
    elif isinstance(nc_paths, list):
        nc_paths1 = nc_paths

    nc_paths1.sort()

    ## Determine duplicate times
    if len(nc_paths1) > 1:
        time_index_bool = tu.processing.determine_duplicate_times(nc_paths1, time_name)
    else:
        time_index_bool = None

    ## Iterate through files
    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
        futures = []
        for nc_path in nc_paths1:
            f = executor.submit(tu.processing.preprocess_data_structure, nc_path, time_name, x_name, y_name, variables, time_index_bool, projected_coordinates)
            futures.append(f)
        runs = concurrent.futures.wait(futures)

    ## process output
    new_paths = [r.result() for r in runs[0]]
    new_paths1 = []
    for new_path in new_paths:
        new_paths1.extend(new_path)

    new_paths1.sort()

    return new_paths1


def resample_to_wgs84_grids(nc_paths, proj4_crs, order=1, min_val=None, max_val=None, bbox=None, time_name='time', x_name='x', y_name='y', max_workers=4):
    """

    """
    if isinstance(nc_paths, str):
        nc_paths1 = glob.glob(nc_paths)
    elif isinstance(nc_paths, list):
        nc_paths1 = nc_paths

    xr1 = xr.open_dataset(nc_paths1[0])

    ## Get approximate grid resolution
    x1 = xr1[x_name].values
    half_x = len(x1)//2
    x = x1[half_x:(half_x+10)]

    y1 = xr1[y_name].values
    half_y = len(y1)//2
    y = y1[half_y:(half_y+10)]

    xr1.close()
    del xr1

    wgs84 = pyproj.Proj(proj='latlong', datum='WGS84')

    trans = pyproj.Transformer.from_proj(proj4_crs, wgs84)

    lon, lat = trans.transform(x, y)

    grid_res_lat = np.quantile(np.abs(np.diff(lat.T)), 0.5)
    grid_res_lon = np.quantile(np.abs(np.diff(lon.T)), 0.5)
    grid_res = round((grid_res_lon + grid_res_lat)/2, 5)

    ## Iterate through files
    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
        futures = []
        for nc_path in nc_paths1:
            f = executor.submit(tu.processing.resample_to_wgs84_grid, nc_path, proj4_crs, grid_res, order, min_val, max_val, bbox)
            futures.append(f)
        runs = concurrent.futures.wait(futures)

    ## process output
    print('grid resolution/length is: ' + str(grid_res))
    new_paths = [r.result() for r in runs[0]]
    new_paths.sort()

    return new_paths


# def calc_new_parameters():
#     """

#     """
#     dates1 = tu.misc.filter_data_paths(obj_list1.path, date_format, return_dates=True)



def combine_metadata(project, dataset):
    """

    """
    datasets = []
    for d in dataset:
        d1 = copy.deepcopy(d)
        d1.update(project)
        datasets.append(d1)

    return datasets


def multi_calc_new_variables(nc_paths, datasets, func_dict):
    """

    """
    dates1 = tu.misc.filter_data_paths(nc_paths, '%Y%m%d%H%M%S', return_dates=True)
    dates1['variable'] = dates1.path.apply(lambda x: pathlib.Path(x).stem.split('_wgs84_')[0])

    ## Iterate through files
    # if max_workers <=1:
    new_paths_list = []
    for i, g in dates1.groupby('date'):
        # print(g)
        p = tu.processing.calc_new_variables(g['path'].tolist(), datasets, func_dict)
        new_paths_list.append(p)

    # elif max_workers > 1:
    #     with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
    #         futures = []
    #         for i, g in dates1.groupby('date'):
    #             # print(g)
    #             f = executor.submit(tu.processing.calc_new_variables, g['path'].tolist(), datasets, func_dict)
    #             futures.append(f)
    #         runs = concurrent.futures.wait(futures)

    #     # process output
    #     new_paths_list = [r.result() for r in runs[0]]

    new_paths1 = []
    for new_path in new_paths_list:
        new_paths1.extend(new_path)

    new_paths1.sort()

    return new_paths1


def multi_mergetime_nc_remove(source_paths, by, max_workers=3):
    """

    """
    source_paths_dict = {}
    if by == 'dataset':
        for p in source_paths:
            path1 = pathlib.Path(p)
            ds_id = path1.stem.split('_')[0]
            if ds_id in source_paths_dict:
                source_paths_dict[ds_id].append(p)
            else:
                source_paths_dict[ds_id] = [p]
    elif by == 'station':
        for p in source_paths:
            path1 = pathlib.Path(p)
            ds_id, stn_id, date = path1.stem.split('_')
            key = (ds_id, stn_id)
            if key in source_paths_dict:
                source_paths_dict[key].append(p)
            else:
                source_paths_dict[key] = [p]
    else:
        raise ValueError('by must be either dataset or station.')

    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
        futures = []
        for ds_id, paths in source_paths_dict.items():
            f = executor.submit(tu.data_io.mergetime_nc_remove, paths, by)
            futures.append(f)
        runs = concurrent.futures.wait(futures)

    new_paths = [r.result() for r in runs[0]]
    new_paths.sort()

    return new_paths







###############################################
### Class


class Titan(object):
    """

    """

    def __init__(self, temp_path: pathlib.Path, run_id: str = None, key: str = None, diagnostics_url: HttpUrl = None ):
        """

        """
        ## Check temp path
        os.makedirs(temp_path, exist_ok=True)

        if not os.access(temp_path, os.W_OK):
            raise OSError('{} has no write access.'.format(temp_path))

        ## Add in the additional paths
        preprocessed_path = os.path.join(temp_path, preprocessed_dir)
        # previous_path = os.path.join(temp_path, previous_dir)
        # final_path = os.path.join(temp_path, final_dir)
        os.makedirs(preprocessed_path, exist_ok=True)
        # os.makedirs(previous_path, exist_ok=True)
        # os.makedirs(final_path, exist_ok=True)

        self.preprocessed_path = preprocessed_path
        # self.previous_path = previous_path
        # self.final_path = final_path

        ## Test for the run_id and key
        need_dgn = True

        if isinstance(key, str) and (diagnostics_url is not None):
            # Attempt to read the diagnostics file
            try:
                resp = requests.get(diagnostics_url)
                f = Fernet(key)
                dgn = read_json_zstd(f.decrypt(resp.content))
                run_id = dgn['run_id']
                need_dgn = False

                _ = [setattr(self, k, v) for k, v in dgn['attributes'].items()]
                print('Diagnostics file has loaded sucessfully.')
            except:
                print('Reading diagnostics file failed. Check the URL link. If you continue, the diagnostics file will be overwritten.')

        if (not isinstance(key, str)):
            key = Fernet.generate_key().decode()
            print('A new encryption key has been generated. Keep it safe and secret:')
            print(key)

        if isinstance(run_id, str):
            if len(run_id) != 14:
                print('The run_id provided is not correct, setting a new one.')
                run_id = uuid.uuid4().hex[:14]
                print('A new run_id has been generated:')
                print(run_id)
        else:
            run_id = uuid.uuid4().hex[:14]
            print('A new run_id has been generated:')
            print(run_id)

        self.key = key
        self.run_id = run_id
        self.temp_path = temp_path

        ## Set up diagnostics dict
        if need_dgn:
            run_date1 = tu.misc.make_run_date_key()
            dgn = {'attributes': {},
                   'run_id': run_id,
                   'run_date': run_date1,
                   'key': key
                   }

        self.diagnostics = dgn

        pass


    @staticmethod
    def process_sparse_stations_from_df(stns, dataset_id=None, connection_config=None, bucket=None, system_version=4):
        """
        Function that takes a stns dataframe of station data and converts it to an Xarray Dataset for Tethys. This is ultimately meant to be combined with the time series data for futher processing. If a geometry column is provided, it must be as a geojson-type dict (not a geopandas column).

        """
        stns = tu.processing.process_sparse_stations_from_df(stns, dataset_id, connection_config=connection_config, bucket=bucket, system_version=system_version)

        return stns


    @staticmethod
    def combine_dataset_metadata(project, dataset):
        """

        """
        datasets = []
        for d in dataset:
            d1 = copy.deepcopy(d)
            d1.update(project)
            datasets.append(d1)

        return datasets


    # def init_results_dict(self):
    #     """

    #     """
    #     results_dict = {d: [] for d in self.datasets}
    #     setattr(self, 'results_dict', results_dict)


    def status_checks(self, connection_config: tdm.base.ConnectionConfig, bucket: str, public_url: str = None, system_version: int = 4, load_diagnostics: bool = False):
        """

        """
        ## Test S3 connection
        _ = tdm.base.Remote(**{'connection_config': connection_config, 'bucket': bucket, 'version': system_version})
        s3 = s3_client(connection_config)

        # Read permissions
        try:
            _ = s3.head_bucket(Bucket=bucket)
        except Exception as err:
            response_code = err.response['Error']['Code']
            if response_code == '403':
                raise requests.exceptions.ConnectionError('403 error. The connection_config is probably wrong.')
            elif response_code == '404':
                raise requests.exceptions.ConnectionError('404 error. The bucket was not found.')

        # Write permissions
        test1 = b'write test'
        test_key = 'tethys/testing/test_write.txt'

        try:
            _ = tu.s3.put_object_s3(s3, bucket, test_key, test1, {}, '', retries=1)
        except bc_exceptions.ConnectionClosedError:
            raise ValueError('Account does not have write permissions.')

        # Public URL read test
        if isinstance(public_url, str):
            try:
                _ = tdm.base.Remote(**{'public_url': public_url, 'bucket': bucket, 'version': system_version})
                _ = get_object_s3(test_key, public_url=public_url, bucket=bucket, counter=1)
            except:
                raise ValueError('public_url does not work.')

        ## Check that the S3 version is valid
        if not system_version in tdm.key_patterns:
            raise ValueError('S3 version must be one of {}.'.format(list(tdm.key_patterns.keys())))

        ## Check if there is an existing Tethys S3 version
        # Version 3+ check
        objs1 = tu.s3.list_objects_s3(s3, bucket, 'tethys/', delimiter='/')
        if not objs1.empty:
            exist_version = int(objs1.iloc[0].Key.split('/')[1].split('.')[0][1:])
        else:
        # Version 2 check
            objs2 = tu.s3.list_objects_s3(s3, bucket, 'tethys/v2/', delimiter='/')
            if objs2.empty:
                exist_version = None
            else:
                exist_version = 2

        if isinstance(exist_version, int):
            if exist_version != system_version:
                print('The bucket already has Tethys data and is version {}. You have been warned...'.format(exist_version))

        print('All checks passed!')

        ## Save parameters
        setattr(self, 'connection_config', connection_config)
        setattr(self, 'bucket', bucket)
        setattr(self, 'public_url', public_url)
        setattr(self, 'system_version', system_version)

        ## Loading diagnostic log if it exists
        if load_diagnostics:
            dgn_key_pattern = tdm.key_patterns[self.system_version]['diagnostics']
            dgn_key = dgn_key_pattern.format(run_id=self.run_id)

            try:
                resp = get_object_s3(dgn_key, bucket, s3, public_url=public_url, counter=0)
                f = Fernet(self.key)
                dgn = read_json_zstd(f.decrypt(resp))
                self.diagnostics = dgn
                _ = [setattr(self, k, v) for k, v in dgn['attributes'].items()]
                print('Diagnostics log found and loaded.')
            except:
                ## diagnostic log
                self.diagnostics['status_checks'] = {'pass': True}
                self.diagnostics['attributes'].update({'connection_config': connection_config, 'bucket': bucket, 'public_url': public_url, 'system_version': system_version})
        else:
            ## diagnostic log
            self.diagnostics['status_checks'] = {'pass': True}
            self.diagnostics['attributes'].update({'connection_config': connection_config, 'bucket': bucket, 'public_url': public_url, 'system_version': system_version})


    def save_diagnostics(self):
        """

        """
        tu.misc.diagnostic_check(self.diagnostics, 'status_checks')

        f = Fernet(self.key)
        dgn_obj = f.encrypt(tu.misc.write_json_zstd(copy.deepcopy(self.diagnostics)))

        dgn_key_pattern = tdm.key_patterns[self.system_version]['diagnostics']
        dgn_key = dgn_key_pattern.format(run_id=self.run_id)

        s3 = s3_client(self.connection_config)
        run_date1 = tu.misc.make_run_date_key()

        obj_resp = tu.s3.put_object_s3(s3, self.bucket, dgn_key, dgn_obj, {'run_date': run_date1}, 'application/json')

        self.diagnostics['diagnostic_s3_obj_id'] = obj_resp['VersionId']

        if isinstance(self.public_url, str):
            dgn_url = b2_public_key_pattern.format(base_url=self.public_url.rstrip('/'), bucket=self.bucket, obj_key=dgn_key)
        else:
            dgn_url = ''

        return dgn_url


    def load_dataset_metadata(self, datasets: Union[dict, list]):
        """

        """
        tu.misc.diagnostic_check(self.diagnostics, 'status_checks')

        dataset_list = tu.processing.process_datasets(datasets)

        # TODO: Check for existing datasets in other S3 buckets

        ds_dict = {ds['dataset_id']: ds for ds in dataset_list}

        setattr(self, 'dataset_list', dataset_list)
        setattr(self, 'datasets', ds_dict)

        ## Check if the chunking parameters are in the datasets
        chunk_params_bool_list = []
        for ds in dataset_list:
            if 'chunk_parameters' in ds:
                _ = tdm.dataset.ChunkParams(**ds['chunk_parameters'])
                chunk_params_bool_list.extend([True])
            else:
                chunk_params_bool_list.extend([True])

        if all(chunk_params_bool_list):
            self.diagnostics['set_chunking_parameters'] = {'pass': True}

        ## diagnostic log
        self.diagnostics['load_dataset_metadata'] = {'pass': True}
        self.diagnostics['attributes'].update({'dataset_list': dataset_list, 'datasets': ds_dict})


    def set_chunking_parameters(self, geometry_count=1, time_interval=None, height_interval=None):
        """

        """
        tu.misc.diagnostic_check(self.diagnostics, 'load_dataset_metadata')

        chunk_params = tdm.dataset.ChunkParams(geometry_count=geometry_count, time_interval=time_interval, height_interval=height_interval).dict(exclude_none=True)

        datasets = []
        for ds in self.dataset_list:
            ds['chunk_parameters'] = chunk_params
            datasets.append(ds)

        self.load_dataset_metadata(datasets)


    def process_versions(self, version_data: dict = None):
        """

        """
        tu.misc.diagnostic_check(self.diagnostics, 'load_dataset_metadata')

        if isinstance(self.public_url, str):
            conn_config = self.public_url
        else:
            conn_config = self.connection_config

        version_dict = tu.s3.process_results_versions(self.dataset_list, bucket=self.bucket, connection_config=self.connection_config, public_url=self.public_url, version_data=version_data, system_version=self.system_version)
        max_version_date_key = tu.misc.make_run_date_key(max(list([d['version_date'] for d in version_dict.values()])))

        # setattr(self, 'processing_code', processing_code)
        setattr(self, 'version_dict', version_dict)
        setattr(self, 'max_version_date_key', max_version_date_key)

        ## diagnostic log
        self.diagnostics['process_versions'] = {'pass': True}
        self.diagnostics['attributes'].update({'version_dict': version_dict, 'max_version_date_key': max_version_date_key})

        return version_dict


    def save_preprocessed_results(self, dataset_id: str, data: Union[List[xr.Dataset], xr.Dataset]):
        """
        Method to save preprocessed results xarray datasets to netcdf files in the temp_path. The results must have the standard dimensions, geometry caclulated, appropriate parameter names, and station data.
        """
        ## All the checks...
        tu.misc.diagnostic_check(self.diagnostics, 'load_dataset_metadata')

        dataset1 = self.datasets[dataset_id]
        result_type = dataset1['result_type']
        param = dataset1['parameter']
        data_model = tdm.dataset.result_type_dict[result_type]
        dims = set(data_model.schema()['properties'].keys())

        if isinstance(data, xr.Dataset):
            data_list = [data]
        else:
            data_list = data

        _ = [data_model(**d.dims) for d in data_list]

        for d in data_list:
            if not param in d:
                raise ValueError('The {param} valiable should be in the data if {ds_id} is the dataset_id.'.format(param=param, ds_id=dataset_id))

            d_dims = set(d[param].dims)
            if d_dims != dims:
                raise ValueError('The {param} valiable should contain the dims: {dims}.'.format(param=param, dims=dims))

        ## Save the data
        file_list = []
        for d in data_list:
            file_id = uuid.uuid4().hex[:14]
            file_path = os.path.join(self.preprocessed_path, preprocessed_file_str.format(ds_id=dataset_id, file_id=file_id))
            _ = tu.misc.write_pkl_zstd(d.to_netcdf(), file_path)
            file_list.append(file_path)

        return file_list


    def get_obj_list(self, glob_path, date_format=None, freq=None, from_date=None, to_date=None, source_connection_config=None, source_bucket=None, source_public_url=None):
        """

        """
        obj_list = get_obj_list(glob_path, date_format=date_format, freq=freq, from_date=from_date, to_date=to_date, connection_config=source_connection_config, bucket=source_bucket)

        # self.obj_list = obj_list.path.tolist()
        self.source_connection_config = source_connection_config
        self.source_bucket = source_bucket
        self.source_public_url = source_public_url

        ## diagnostic log
        self.diagnostics['get_obj_list'] = {'pass': True}
        self.diagnostics['attributes'].update({'source_connection_config': source_connection_config, 'source_bucket': source_bucket, 'source_public_url': source_public_url})

        return obj_list


    def copy_source_objs(self, source_paths, compression=None, threads=None, max_workers=None):
        """

        """
        tu.misc.diagnostic_check(self.diagnostics, 'get_obj_list')

        new_paths = copy_source_objs(source_paths, self.preprocessed_path, threads=threads, compression=compression, connection_config=self.source_connection_config, bucket=self.source_bucket, public_url=self.source_public_url, max_workers=max_workers)

        return new_paths


    # def decompress_files(self, source_paths, compression_type='gzip', max_workers=4):
    #     """

    #     """
    #     new_paths1 = decompress_path(source_paths, compression_type=compression_type, max_workers=max_workers)

    #     return new_paths1


    def file_format_conversion(self, source_paths, file_format='grib', max_workers=4):
        """

        """
        new_paths2 = file_format_conversion(source_paths, file_format=file_format, max_workers=max_workers)

        return new_paths2


    def variable_processing(self, source_paths, time_name, x_name, y_name, variables, projected_coordinates=True, nc_source='normal', max_workers=4):
        """

        """
        new_paths2 = variable_processing(source_paths, time_name, x_name, y_name, variables, projected_coordinates=projected_coordinates, max_workers=max_workers)

        return new_paths2


    def resample_to_wgs84(self, source_paths, proj4_crs, order=2, bbox=None, max_workers=4):
        """

        """
        new_paths2 = resample_to_wgs84_grids(source_paths, proj4_crs, order=order, bbox=bbox, max_workers=max_workers)

        return new_paths2


    def calc_new_variables(self, source_paths, func_dict):
        """

        """
        new_paths2 = multi_calc_new_variables(source_paths, self.dataset_list, func_dict)

        return new_paths2


    def merge_time_nc_files(self, source_paths, by, max_workers=4):
        """

        """
        if by == 'station':
            tu.misc.diagnostic_check(self.diagnostics, 'copy_interim_objs')

        new_paths = multi_mergetime_nc_remove(source_paths, by=by, max_workers=max_workers)

        if by == 'station':
            ## diagnostic log
            self.diagnostics['merge_nc_files'] = {'pass': True}

        return new_paths


    def save_dataset_stations(self, source_paths, block_length, block_length_factor=10, compression='zstd', max_workers=4):
        """

        """
        new_paths = multi_save_dataset_stations(source_paths, block_length, block_length_factor=block_length_factor, compression=compression, max_workers=max_workers)

        return new_paths


    def upload_interim_results(self, source_paths, threads=10):
        """

        """
        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
            s3 = s3_client(self.connection_config, max_pool_connections=threads)
            futures = []
            for path in source_paths:
                f = executor.submit(tu.s3.put_interim_results_s3, s3, self.bucket, path, self.run_id, system_version=self.system_version)
                futures.append(f)
            runs = concurrent.futures.wait(futures)

        keys = [r.result() for r in runs[0]]

        self.interim_keys = keys

        ## Remove source files
        for path in source_paths:
            os.remove(path)

        ## diagnostic log
        # self.diagnostics['upload_interim_results'] = {'pass': True}
        # self.diagnostics['attributes'].update({'interim_keys': keys})

        return keys


    def get_interim_results_list(self, max_size_gb=0.1):
        """

        """
        tu.misc.diagnostic_check(self.diagnostics, 'status_checks')

        key_pattern = tdm.key_patterns[self.system_version]['interim_results'].split('{dataset_id}')[0]
        key = key_pattern.format(run_id=self.run_id)

        s3 = s3_client(self.connection_config)
        obj_list1 = tu.s3.list_objects_s3(s3, self.bucket, key)
        obj_list1 = obj_list1[['Key', 'Size']].rename(columns={'Key': 'key', 'Size': 'size'})

        obj_keys1 = obj_list1['key'].tolist()
        ds_keys2 = [k.split('/')[4] for k in obj_keys1]
        stn_keys2 = [k.split('/')[5].split('.')[0] for k in obj_keys1]

        obj_list1['dataset_id'] = ds_keys2
        obj_list1['station_id'] = stn_keys2

        # ds_sizes = obj_list1.groupby('dataset_id')['size'].sum().reset_index()
        ds_stn_sizes = obj_list1.groupby(['dataset_id', 'station_id'])['size'].sum().reset_index()
        ds_stn_sizes = ds_stn_sizes.sample(frac=1, ignore_index=True, random_state=1)

        # self.ds_sizes = ds_sizes
        # self.ds_stn_sizes = ds_stn_sizes
        # self.interim_objects = obj_list1

        max_size = max_size_gb * 1000000000

        ds_stn_sum = ds_stn_sizes['size'].sum()
        mean_size = ds_stn_sum/len(ds_stn_sizes)
        group_len = int(max_size//mean_size)
        df_len = len(ds_stn_sizes)

        grouper = np.repeat(np.arange((df_len//group_len)+1), group_len)

        grp1 = ds_stn_sizes.groupby(grouper[:df_len])
        ds_stn_groups = [v for k, v in grp1]

        interim_groups = []
        for df in ds_stn_groups:
            ig = pd.merge(df.drop('size', axis=1), obj_list1, on=['dataset_id', 'station_id'])
            interim_groups.append(ig)

        ## diagnostic log
        self.diagnostics['get_interim_results_list'] = {'pass': True}
        # self.diagnostics['attributes'].update({'interim_objects': obj_list1, 'ds_stn_sizes': ds_stn_sizes, 'ds_sizes': ds_sizes})

        return interim_groups


    def copy_interim_objs(self, source_paths, threads=20):
        """

        """
        tu.misc.diagnostic_check(self.diagnostics, 'get_interim_results_list')

        new_paths = copy_interim_objs(source_paths, self.preprocessed_path, threads=threads, connection_config=self.connection_config, bucket=self.bucket, public_url=self.public_url)

        ## diagnostic log
        self.diagnostics['copy_interim_objs'] = {'pass': True}

        return new_paths


    def save_final_results(self, source_paths, correct_times=False, max_workers=4):
        """

        """
        tu.misc.diagnostic_check(self.diagnostics, 'merge_nc_files')
        tu.misc.diagnostic_check(self.diagnostics, 'process_versions')

        ## Put the big files first
        merge_list = [[p, os.path.getsize(p)] for p in source_paths]
        merge_df = pd.DataFrame(merge_list, columns=['path', 'size'])

        source_paths1 = merge_df.sort_values('size', ascending=False).path.tolist()

        ## Iterate through files
        with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
            futures = []
            for source_path in source_paths1:
                ds_id = source_path.split('/')[-1].split('_')[0]
                metadata = self.datasets[ds_id]
                f = executor.submit(tu.processing.save_final_results, source_path, metadata, self.max_version_date_key, correct_times, self.system_version)
                futures.append(f)
            runs = concurrent.futures.wait(futures)

        ## process output
        results_new_paths1 = []
        stn_new_paths1 = []

        new_paths = [r.result() for r in runs[0]]

        for paths_list in new_paths:
            r1, s1 = paths_list
            results_new_paths1.extend(r1)
            stn_new_paths1.extend(s1)

        results_new_paths1.sort()
        stn_new_paths1.sort()

        ## diagnostic log
        self.diagnostics['save_final_results'] = {'pass': True}

        return results_new_paths1, stn_new_paths1


    def upload_final_results(self, results_paths, threads=30):
        """

        """
        tu.misc.diagnostic_check(self.diagnostics, 'save_final_results')

        ## Determine which chunks need to uploaded
        if isinstance(self.public_url, str):
            remote1 = {'bucket': self.bucket, 'public_url': self.public_url, 'version': self.system_version}
        else:
            remote1 = {'bucket': self.bucket, 'connection_config': self.connection_config, 'version': self.system_version}

        r_chunks = tu.s3.determine_results_chunks_diffs(results_paths, remote1)

        if not r_chunks.empty:
            results_paths1 = r_chunks['file_path'].tolist()

            random.shuffle(results_paths1)

            ## Make sure all of the station files exist
            stn_paths = []
            for path in results_paths1:
                ds_id, version_date1, stn_id, chunk_id, hash_id, _ = os.path.split(path)[1].split('_')

                stn_file_name = tu.processing.ds_stn_json_str.format(ds_id=ds_id, stn_id=stn_id)
                stn_file_path = os.path.join(os.path.split(path)[0], stn_file_name)

                if os.path.exists(stn_file_path):
                    stn_paths.append(stn_file_path)
                else:
                    raise FileNotFoundError('station file not found: ' + stn_file_path)

            ## Iterate through the results and stations
            with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
                s3 = s3_client(self.connection_config, max_pool_connections=threads)
                futures = []

                for path in results_paths1:
                    f = executor.submit(tu.s3.put_result, s3, self.bucket, path, system_version=self.system_version)
                    futures.append(f)

                for path in set(stn_paths):
                    # Add the station file
                    f = executor.submit(tu.s3.put_station, s3, self.bucket, path, system_version=self.system_version)
                    futures.append(f)

                runs = concurrent.futures.wait(futures)

            keys = [r.result() for r in runs[0]]
            keys.sort()

            stn_keys = [key for key in keys if 'station' in key]
            results_keys = [key for key in keys if 'results' in key]
        else:
            stn_keys = []
            results_keys = []
            print('No chunks have changed, so no files have been uploaded.')

        ## Remove source files
        path1 = pathlib.Path(os.path.split(results_paths[0])[0])
        for path in path1.glob('*_results.nc.zst'):
            os.remove(path)
        for path in path1.glob('*_station.json.zst'):
            os.remove(path)

        return results_keys, stn_keys


    def update_aggregates(self, threads=60):
        """

        """
        ## Update the datasets and station jsons
        s3 = s3_client(self.connection_config, threads)

        for ds in self.dataset_list:
            dataset_id = ds['dataset_id']

            print(dataset_id)

            print('-- Aggregating stations')
            stns_list = tu.s3.put_remote_agg_stations(dataset_id, self.bucket, s3, self.public_url, threads=threads, system_version=self.system_version)

            if stns_list:
                print('-- Updating versions')
                version_data1 = self.version_dict[dataset_id]

                version_list =  tu.s3.update_versions(version_data1, self.bucket, self.connection_config, self.public_url, system_version=self.system_version)

                print('-- Updating Dataset')
                ds_new = tu.s3.put_dataset(ds, self.bucket, s3=s3, public_url=self.public_url, system_version=self.system_version)

        # Aggregate all datasets for the bucket
        print('-- Aggregating datasets')
        ds_all = tu.s3.put_remote_agg_datasets(self.bucket, s3, self.public_url, 30, system_version=self.system_version)

        return ds_all


    def clear_interim_results(self):
        """

        """
        prefix = tdm.utils.key_patterns[self.system_version]['interim_results'].split('{dataset_id}')[0].format(run_id=self.run_id)

        s3 = s3_client(self.connection_config)

        obj_list = tu.s3.list_object_versions_s3(s3, self.bucket, prefix)

        rem_keys = []
        for i, row in obj_list.iterrows():
            rem_keys.extend([{'Key': row['Key'], 'VersionId': row['VersionId']}])

        if len(rem_keys) > 0:
            ## Split them into 1000 key chunks
            rem_keys_chunks = np.array_split(rem_keys, int(np.ceil(len(rem_keys)/1000)))

            ## Run through and delete the objects...
            for keys in rem_keys_chunks:
                _ = s3.delete_objects(Bucket=self.bucket, Delete={'Objects': keys.tolist(), 'Quiet': True})

        print(str(len(rem_keys)) + ' objects removed')












################################
### Testing


   # def init_ray(self, num_cpus=1, include_dashboard=False, configure_logging=False, **kwargs):
    #     """

    #     """
    #     if ray.is_initialized():
    #         ray.shutdown()

    #     ray.init(num_cpus=num_cpus, include_dashboard=include_dashboard, configure_logging=configure_logging, **kwargs)

    #     @ray.remote
    #     def _load_result(dataset, result, run_date_key, other_attrs, discrete, other_closed, sum_closed, other_encoding):
    #         """

    #         """
    #         out1 = tu.processing.prepare_results_v02(dataset, result, run_date_key, sum_closed=sum_closed, other_closed=other_closed, discrete=discrete, other_attrs=other_attrs, other_encoding=other_encoding)

    #         return out1

    #     self._load_result = _load_result
    #     self._obj_refs = []


    # def shutdown_ray(self):
    #     ray.shutdown()


    # def load_results(self, results, sum_closed='right', other_closed='left', discrete=True, other_attrs=None, other_encoding=None, run_date=None):
    #     """

    #     """
    #     ## Dataset checks
    #     # ds_ids = list(results.keys())

    #     if isinstance(run_date, str):
    #         run_date_key = tu.misc.make_run_date_key(run_date)
    #     else:
    #         run_date_key = self.max_run_date_key

    #     r1 = [self._load_result.remote(self.datasets[r['dataset_id']], r['result'], run_date_key, other_attrs, discrete, other_closed, sum_closed, other_encoding) for r in results]
    #     # r2 = ray.get(r1)

    #     self._obj_refs.extend(r1)



# -*- coding: utf-8 -*-
"""
Created on Tue Jul  2 09:25:41 2019

@author: michaelek
"""
import pytest
from tethys_utils import *
from tethysts import Tethys
import pandas as pd
# from tethys_utils.datasets import get_path

pd.options.display.max_columns = 10


###############################################
### Parameters

# remote = {'bucket': 'nz-open-modelling-consortium', 'connection_config': 'https://b2.nzrivers.xyz', 'version': 3}
remote = {'bucket': 'ecan-env-monitoring', 'connection_config': 'https://b2.tethys-ts.xyz', 'version': 2}

attrs = {'quality_code': {'standard_name': 'quality_flag',
  'long_name': 'NEMS quality code',
  'references': 'https://www.lawa.org.nz/media/16580/nems-quality-code-schema-2013-06-1-.pdf'},
  'well_depth': {'units': 'm'},
  'well_diameter': {'units': 'mm'},
  'well_screens': {'units': ''},
  'well_top_screen': {'units': 'm'},
  'well_bottom_screen': {'units': 'm'},
  }

encoding = {'quality_code': {'dtype': 'int16', '_FillValue': -9999},
  'well_depth': {'dtype': 'int32', '_FillValue': -99999, 'scale_factor': 0.1},
  'well_diameter': {'dtype': 'int32',
   '_FillValue': -99999,
   'scale_factor': 0.1},
  'well_screens': {'dtype': 'int16', '_FillValue': -9999},
  'well_top_screen': {'dtype': 'int32',
   '_FillValue': -99999,
   'scale_factor': 0.1},
  'well_bottom_screen': {'dtype': 'int32',
   '_FillValue': -99999,
   'scale_factor': 0.1},
  'groundwater_depth': {'scale_factor': 0.001,
     'dtype': 'int32',
     '_FillValue': 9999}}

ds_id = 'f98dfaefbc15f045f900e4b9'
station_id = 'fff111e6e8e652b0804a21b4'
parameter = 'groundwater_depth'

########################################
### Tests

self = Tethys([remote])

stns = self.get_stations(ds_id)

results = self.get_results(ds_id, station_id)
data = results.drop(['lat', 'lon'])

param_attrs = data[parameter].attrs.copy()
_ = [param_attrs.pop(p) for p, v in data[parameter].attrs.items() if p in ['dataset_id', 'ancillary_variables', 'result_type']]
param_attrs['spatial_distribution'] = 'sparse'
param_attrs['geometry_type'] = 'Point'
param_attrs['grouping'] = 'none'
param_attrs['parent_datasets'] = ['rrrrrrr', 'ttttttt']

attrs[parameter] = param_attrs

res3 = package_xarray(data, parameter, attrs, encoding)



def test_read_pkl_zstd():
    df1 = read_pkl_zstd(d_path1)

    assert df1.shape == (20000, 7)



def test_write_pkl_zstd():
    p_df1 = write_pkl_zstd(df1)
    len1 = round(len(p_df1), -3)

    assert (len1 < 200000) and (len1 > 100000)


def test_df_to_xarray():
    p_ds1 = df_to_xarray(df1, nc_type, param_name, attrs, encoding, run_date_key, ancillary_variables, compression)
    len2 = round(len(p_ds1), -3)

    ds1 = df_to_xarray(df1, nc_type, param_name, attrs, encoding, run_date_key, ancillary_variables)

    assert (len(ds1) == 6) and (len2 < 30000) and (len2 > 20000)




# @pytest.mark.parametrize('input_sites', [input_sites1, input_sites2, input_sites3])
# def test_nat(input_sites):
#     f1 = FlowNat(from_date, to_date, input_sites=input_sites)
#
#     nat_flow = f1.naturalisation()
#
#     assert (len(f1.summ) >= 1) & (len(nat_flow) > 2900)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 20 10:43:15 2021

@author: mike
"""
from io import BytesIO
import xarray as xr
import os
import gzip
import shutil
import requests
import orjson
import copy
import multiprocessing as mp
# from cdo import Cdo
import pathlib
from tethysts.utils import get_object_s3, url_stream_to_file, read_pkl_zstd, b2_public_key_pattern
import gzip
import zstandard as zstd
import subprocess
# mp.set_start_method('spawn')

##########################################
### Functions


def query_esri_mapserver(base_url, layer_id, out_fields=None, where=None):
    """
    Query an ESRI map server for a vector layer and return a geojson structured dict.

    Parameters
    ----------
    base_url : str
        The base url for the map server up to the layer_id.
    layer_id : str or int
        The layer id.
    out_fields : None or list of str
        The output fields to be returned. The geometry will always be returned. None will return all fields.
    where : None or str
        The SQL style 'where' clause to query the layer. None will have no filters.

    Returns
    -------
    Dict
        In GeoJSON structure
    """
    url = base_url + str(layer_id) + '/query'

    ## Set up filters
    params = {'returnGeometry': 'true', 'f': 'geojson'}

    if out_fields is None:
        params.update({'outFields': '*'})
    elif isinstance(out_fields, list):
        out_str = ', '.join(out_fields)
        params.update({'outFields': out_str})
    else:
        raise ValueError('out_fields must either be None or a list of strings')

    if where is None:
        params.update({'where': '1=1'})
    elif isinstance(where, str):
        params.update({'where': where})
    else:
        raise ValueError('where must either be None or a string')

    ## Run the queries
    resp = requests.get(url, params=params, timeout=300)

    if not resp.ok:
        raise ValueError(resp.raise_for_status())

    data = orjson.loads(resp.content)

    if 'exceededTransferLimit' in data:
        exceed = data.pop('exceededTransferLimit')
    else:
        exceed = False

    data_all = copy.deepcopy(data)

    while exceed:
        last_id = data['features'][-1]['id']

        if isinstance(where, str):
            params.update({'where': where + 'and objectid > ' + str(last_id)})
        else:
            params.update({'where': 'objectid > ' + str(last_id)})

        resp = requests.get(url, params=params, timeout=300)

        if not resp.ok:
            raise ValueError(resp.raise_for_status())

        data = orjson.loads(resp.content)

        data_all['features'].extend(data['features'])

        if 'exceededTransferLimit' in data:
            exceed = data.pop('exceededTransferLimit')
        else:
            exceed = False

    return data_all



def decompress_file(source_path, dest_path=None, compression_type='gzip', buffer_size=5242880):
    if not isinstance(dest_path, str):
        dest_path = os.path.splitext(source_path)[0]
    with gzip.open(source_path, 'rb') as s_file, open(dest_path, 'wb') as d_file:
        shutil.copyfileobj(s_file, d_file, buffer_size)

    return dest_path


def dfd(source_path):
    dest_path = decompress_file(source_path)
    os.remove(source_path)
    return dest_path


def decompress_files_remove(source_paths, max_workers=4):
    """

    """

    # with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers) as executor:
    #     results = executor.map(dfd, source_paths)
    with mp.get_context("spawn").Pool(processes=max_workers) as pool:
        results = pool.map(dfd, source_paths)

    return results


def convert_grib_to_nc(source_path):
    """

    """
    dest_path = os.path.splitext(source_path)[0]
    dest_path = dest_path + '.nc'
    cmd_str = 'cdo -s -O -f nc4 copy {in_file} {out_file}'.format(in_file=source_path, out_file=dest_path)

    _ = subprocess.run(cmd_str, shell=True, check=True)

    # cdo = Cdo()
    # cdo.copy(input=source_path, output=dest_path, options = "-O -f nc4")

    return dest_path


def convert_grib_remove(source_path):
    dest_path = convert_grib_to_nc(source_path)
    os.remove(source_path)
    return dest_path


def convert_gribs_remove(source_paths, max_workers=4):
    """

    """
    with mp.get_context("spawn").Pool(processes=max_workers) as pool:
        results = pool.map(convert_grib_remove, source_paths)

    return results


# def copy_file(source_path, dest_path, buffer_size=10485760):
#     dest_path1 = pathlib.Path(dest_path)
#     if dest_path1.is_dir():
#         source_path1 = pathlib.Path(source_path)
#         file_name = source_path1.name
#         dest_path2 = str(dest_path1.joinpath(file_name))
#     else:
#         dest_path2 = dest_path
#     with open(source_path, 'rb') as s_file, open(dest_path2, 'wb') as d_file:
#         shutil.copyfileobj(s_file, d_file, buffer_size)

#     return dest_path2


def copy_file(source_path, dest_path, compression=None, chunk_size=524288):
    dest_path1 = pathlib.Path(dest_path)
    if dest_path1.is_dir():
        source_path1 = pathlib.Path(source_path)
        file_name = source_path1.name
        dest_path2 = str(dest_path1.joinpath(file_name))
    else:
        dest_path2 = dest_path

    if compression == 'gzip':
        if dest_path2.endswith('.gz'):
            dest_path2 = os.path.splitext(dest_path2)[0]
        with gzip.open(source_path, 'rb') as s_file, open(dest_path2, 'wb') as d_file:
            shutil.copyfileobj(s_file, d_file, chunk_size)
    elif compression == 'zstd':
        if dest_path2.endswith('.zst'):
            dest_path2 = os.path.splitext(dest_path2)[0]
        with open(source_path, 'rb') as s_file, open(dest_path2, 'wb') as d_file:
            dctx = zstd.ZstdDecompressor()
            dctx.copy_stream(s_file, d_file, read_size=chunk_size, write_size=chunk_size)
    else:
        shutil.copyfile(source_path, dest_path2)

    return dest_path2


def mergetime_nc_files(nc_paths, by):
    """

    """
    path1 = pathlib.Path(nc_paths[0])
    parent_path = path1.parent

    if by == 'dataset':
        ds_id = path1.stem.split('_')[0]
        file_name = ds_id + '.nc'
    elif by == 'station':
        ds_id, stn_id, date = path1.stem.split('_')
        file_name = '{}_{}.nc'.format(ds_id, stn_id)

    dest_path = str(parent_path.joinpath(file_name))
    source_paths = ' '.join(nc_paths)

    cmd_str = 'cdo -s -O mergetime {in_files} {out_file}'.format(in_files=source_paths, out_file=dest_path)

    _ = subprocess.run(cmd_str, shell=True, check=True)

    # cdo = Cdo()
    # _ = cdo.mergetime(input=source_paths, output=dest_path)

    return dest_path


def mergetime_nc_remove(source_paths, by):
    dest_path = mergetime_nc_files(source_paths, by)
    for p in source_paths:
        os.remove(p)
    return dest_path


def copy_s3_file(source_path, dest_path, bucket, s3=None, connection_config=None, public_url=None, compression=None, chunk_size=524288):
    """

    """
    file_path1 = pathlib.Path(dest_path)
    if file_path1.is_dir():
        file_name = source_path.split('/')[-1]
        file_path2 = str(file_path1.joinpath(file_name))
    else:
        file_path2 = dest_path

    base_path = os.path.split(file_path2)[0]
    os.makedirs(base_path, exist_ok=True)

    counter = 5
    while True:
        if counter < 1:
            raise gzip.BadGzipFile('Too many crc32 errors for: ' + source_path)
        try:
            if isinstance(public_url, str):
                url1 = b2_public_key_pattern.format(base_url=public_url.rstrip('/'), bucket=bucket, obj_key=source_path)
                file_path3 = url_stream_to_file(url1, file_path2, compression=compression, chunk_size=chunk_size)
            else:
                url1 = get_object_s3(source_path, connection_config=connection_config, bucket=bucket, counter=5)
                if compression == 'zstd':
                    if file_path2.endswith('.zst'):
                        file_path2 = os.path.splitext(file_path2)[0]
                    url1 = read_pkl_zstd(url1)
                elif compression == 'gzip':
                    if file_path2.endswith('.gz'):
                        file_path2 = os.path.splitext(file_path2)[0]
                    url1 = gzip.decompress(url1)
                b1 = BytesIO(url1)
                with open(file_path2, 'wb') as f:
                    chunk = b1.read(chunk_size)
                    while chunk:
                        f.write(chunk)
                        chunk = b1.read(chunk_size)

                file_path3 = file_path2
            break
        except gzip.BadGzipFile:
            print('crc32 gzip error for: ' + source_path)
            counter = counter - 1

    return file_path3



# -*- coding: utf-8 -*-
"""
Created on Tue Jul  2 09:25:41 2019

@author: michaelek
"""
import glob
import numpy as np
import os
import yaml
# import pytest
import xarray as xr
import pandas as pd
# from tethys_cdsapi import Processor
import concurrent.futures
from time import time
import tethys_utils as tu
import requests
from tethysts import utils

pd.options.display.max_columns = 10


###############################################
### Parameters

base_dir = os.path.realpath(os.path.dirname(__file__))

with open(os.path.join(base_dir, 'parameters1.yml')) as param:
    param = yaml.safe_load(param)

source = param['source'].copy()
remote = param['remote']['s3']
public_url = source['public_url']
processing_code = source['processing_code']
connection_config = remote['connection_config']
bucket = remote['bucket']

datasets = source['datasets']
# parameter = datasets[0]['parameter']

data_path1 = '/media/sdb1/Data/ecmwf/era5-land/2m_temperature_2003-2013_reanalysis-era5-land.nc'
# data_path1 = '/media/sdb1/Data/ecmwf/era5-land/2m_temperature_2014-2020_reanalysis-era5-land.nc'
data_path2 = '/media/sdb1/Data/ecmwf/era5-land/2m_temperature_*.nc'
data_path4 = '/media/sdb1/Data/ecmwf/era5-land/2m_temperature_2003-2013_reanalysis-era5-land_*.nc'
data_path3 = '/media/sdb1/Data/ecmwf/era5-land/2m_temperature-test1.nc'
data_path4 = '/media/sdb1/Data/ecmwf/era5-land/test/*.nc'
data_path5 = '/media/sdb1/Data/ecmwf/era5-land/test/total_evaporation_1981-1991_reanalysis-era5-land.nc'
data_path = ['/media/sdb1/Data/ecmwf/era5-land/test/2m_temperature_*.nc', '/media/sdb1/Data/ecmwf/era5-land/test/total_evaporation_*.nc']
parameter_code = 'temp_at_2'

grid = data_path2

encoding = {'dtype': 'int16', '_FillValue': -9999, 'scale_factor': 0.01}

lat_dim_name = 'latitude'
lon_dim_name = 'longitude'
time_dim_name = 'time'
dataset_id = 'b2f6bdd8aa592dacb3b257c0'
samples = 900

########################################
### Tests


# def preprocessor(ds):
#     """

#     """
#     ds = ds.rename({'longitude': 'lon', 'latitude': 'lat', 't2m': 'temperature'})
#     ds = ds.assign_coords({'height': 2})
#     ds = ds.expand_dims('height')
#     ds['temperature'] = ds['temperature'] - 273.15
#     ds['temperature'].encoding = encoding

#     return ds

# p1 = Processor(data_path, parameter_code, False)
# ds1 = p1.build_dataset()


# self = Grid(datasets, remote, processing_code, public_url)

# self.load_data(data, parameter, height=2)

# block_test = self.determine_grid_block_size(starting_x_size=20, starting_y_size=20, increment=10, min_size=800, max_size=1100)

# x_size = block_test['x_size']
# y_size = block_test['y_size']

d
# ds = xr.open_dataset(data_path1)
ds = xr.open_mfdataset(data_path2, chunks={'time': 1}, parallel=True, preprocess=preprocessor)
ds = xr.open_mfdataset(data_path2, chunks={'time': 350639, 'lon': 1, 'lat': 1},  preprocess=preprocessor)

ds = xr.open_mfdataset(data_path1, chunks={'time': 1}, parallel=True)
ds = xr.open_mfdataset(data_path2, chunks={'longitude': 20, 'latitude': 20}, parallel=True)

ds = ds.rename({'longitude': 'lon', 'latitude': 'lat', 't2m': 'temperature'})
ds = ds.assign_coords({'height': 2})
ds = ds.expand_dims('height')
ds['temperature'] = ds['temperature'] - 273.15
ds['temperature'].encoding = encoding

da = ds.temperature

da = da.chunk({'lon': 12})

chunks = np.arange(0, 12*(124//12+1)+12, 12)

file_name_str = '{}_part_{:>02d}.nc'

for i, c in enumerate(chunks[:-1]):
    print(c)
    new_file_path = file_name_str.format(os.path.splitext(data_path1)[0], i+1)
    ds1 = da.isel(lon=slice(chunks[i], chunks[i+1]))
    ds1.to_netcdf(new_file_path)


ds = ds.chunk({'time': 1})
ds5 = ds.isel(height=0, time=0, drop=True)

grid = ds

x_size=20
y_size=20
n_intervals = 4
samples=10

# x_name='longitude'
# y_name = 'latitude'

# ds4 = xr.open_dataset(data_path3)

# chunks = {'time': 10, 'latitude': 10, 'longitude': 10}
# chunks = {'time': 10000, 'latitude': 10, 'longitude': 10}
chunks = {'time': 10000}

self = Processor()

ds_list = self.load_dataset_metadata(datasets)

grid1 = self.load_raw_grid(data_path2, lon_dim_name, lat_dim_name, time_dim_name, preprocessor=tu.grid.processors.era5_land.preprocessor)



self.save_processed_grid()


da.to_netcdf(data_path3)
ds.to_netcdf(data_path3)
delayed_obj = ds.to_netcdf(data_path3, compute=False, engine="netcdf4")

with ProgressBar():
    results = delayed_obj.compute()


ds3 = xr.open_mfdataset(data_path4)


nc_files1 = glob.glob(data_path2)
nc_files1.sort()

base_path = os.path.split(data_path2)[0]

for f in nc_files1:
    print(f)
    ds = xr.open_mfdataset(f, chunks={'longitude': 12})

    ds = ds.rename({'longitude': 'lon', 'latitude': 'lat', 't2m': 'temperature'})
    ds = ds.assign_coords({'height': 2})
    ds = ds.expand_dims('height')
    ds['temperature'] = ds['temperature'] - 273.15
    ds['temperature'].encoding = encoding

    chunks = np.arange(0, 12*(124//12+1)+12, 12)

    file_name_str = '{}_part_{:>02d}.nc'

    for i, c in enumerate(chunks[:-1]):
        print(c)
        file_name = os.path.splitext(os.path.split(f)[1])[0]
        new_file_path = file_name_str.format(os.path.join(base_path, 'temp', file_name), i+1)
        ds1 = ds.isel(lon=slice(chunks[i], chunks[i+1]))
        ds1.to_netcdf(new_file_path)
        ds1.close()

    ds.close()
    ds = None



ds3 = xr.open_mfdataset(os.path.join(base_path, 'temp', '*.nc'))

ds3.to_netcdf(data_path3)


nc = netCDF4.Dataset(data_path1)
nc.set_auto_mask(False)

# Change staggered variables to unstaggered ones
for vn, v in nc.variables.items():
    if wrftools.Unstaggerer.can_do(v):
        nc.variables[vn] = wrftools.Unstaggerer(v)

# Check if we can add diagnostic variables to the pot
for vn in wrftools.var_classes:
    cl = getattr(wrftools, vn)
    if vn not in nc.variables and cl.can_do(nc):
        nc.variables[vn] = cl(nc)

# trick xarray with our custom netcdf
ds = xr.open_dataset(NetCDF4DataStore(nc))




ds = xr.open_mfdataset(data_path2)



def save1(b):
    """

    """
    c = b.copy()
    c.load().to_netcdf(data_path3)
    c.close()
    c = None

    return ds



ds6 = xr.map_blocks(save1, ds)



block_list = split_grid(ds, x_size=20, y_size=20, n_intervals=None, x_name='lon', y_name='lat', mbytes_size=1500)

s1 = time()
for i, b in enumerate(block_list[:10]):
    print(i)
    c = b.copy()
    c.load().to_netcdf(data_path3)
    c.close()
    c = None

e1 = time()

diff1 = e1 - s1
print(diff1)



# s2 = time()
# with concurrent.futures.ProcessPoolExecutor(max_workers=2) as executor:
#     results = executor.map(save1, block_list[:10])

# e2 = time()

# diff2 = e2 - s2
# print(diff2)

base_path = os.path.split(data_path2)[0]
output_path = os.path.join(base_path, 'temp')

self = Processor()

ds_list = self.load_dataset_metadata(datasets)

# grid1 = self.load_raw_grid(data_path2, lon_dim_name, lat_dim_name, time_dim_name, preprocessor=tu.grid.processors.era5_land.preprocessor)
grid1 = self.load_raw_grid(data_path4, lon_dim_name, lat_dim_name, time_dim_name, preprocessor=preprocessor)

obj_dict = self.determine_grid_block_size(ds_list[0]['dataset_id'])

# self.save_grid_chunks(output_path)

self.save_grid_chunks(output_path, 1)



self = Grid()

ds_list = self.load_dataset_metadata(datasets)

self.load_connection_params(connection_config, bucket, public_url)

run_date='2021-08-29'

self.load_run_date(processing_code, run_date=run_date)

base_path = os.path.split(data_path2)[0]
output_path = os.path.join(base_path, 'temp')
results = os.path.join(output_path, '*.nc')

xy_size = 1
# max_workers = 2
dataset_id=None
sum_closed='right'
other_closed='left'
discrete=False
other_attrs=None
other_encoding=None

self.load_results(results, xy_size, dataset_id, sum_closed, other_closed, discrete, other_attrs, other_encoding)

# self.public_url = public_url

self.update_results(threads=20)

self.update_aggregates()





url1 = 'https://b2.tethys-ts.xyz/file/tethysts/tethys/v2/b2f6bdd8aa592dacb3b257c0/0dce1a2a943fd46e6e7ce6f4/station.json.zst'


resp = requests.get(url1)

s1 = utils.read_json_zstd(resp.content)





















































































#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat May  1 13:30:43 2021

@author: mike
"""
import os
import numpy as np
import zstandard as zstd
import pandas as pd
import xarray as xr
import orjson
from tethys_data_models.dataset import Dataset, Station
# from data_models import Geometry, Dataset, DatasetBase, S3ObjectKey
import tethys_data_models as tdm
from tethysts.utils import get_object_s3, s3_client, read_json_zstd, read_pkl_zstd, chunk_filters
from tethys_utils.misc import make_run_date_key, write_json_zstd, path_date_parser
# from misc import make_run_date_key, write_json_zstd, write_pkl_zstd
from tethys_utils.processing import compare_xrs, process_station_summ, stats_for_dataset_metadata
# from processing import compare_xrs, process_station_summ, stats_for_dataset_metadata
from datetime import date, datetime
from tethysts import Tethys
from botocore import exceptions as bc_exceptions
from time import sleep
import pathlib
import concurrent.futures
import multiprocessing as mp

############################################
### Parameters



############################################
### Functions


def put_object_s3(s3, bucket, key, obj, metadata, content_type, retries=5):
    """

    """
    counter = retries
    while counter > 0:
        try:
            obj2 = s3.put_object(Bucket=bucket, Key=key, Body=obj, Metadata=metadata, ContentType=content_type)
            break
        except bc_exceptions.ConnectionClosedError as err:
            print(err)
            counter = counter - 1
            if counter == 0:
                raise err
            print('...trying again...')
            sleep(5)
        except bc_exceptions.ConnectTimeoutError as err:
            print(err)
            counter = counter - 1
            if counter == 0:
                raise err
            print('...trying again...')
            sleep(5)

    return obj2


def put_file_s3(s3, bucket, key, file_path, metadata, content_type, retries=5):
    """

    """
    with open(file_path, 'rb') as f:
        obj_out = put_object_s3(s3, bucket, key, f.read(), metadata, content_type, retries=retries)

    return obj_out


def copy_object_s3(s3, source_bucket, source_key, dest_bucket, dest_key, retries=5):
    """
    Copies an object in an S3 database to another location in an S3 database. They must have the same fundemental connection_config. All metadata is copied to the new object.
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.copy_object
    """
    source_dict = {'Bucket': source_bucket, 'Key': source_key}

    counter = retries
    while counter > 0:
        try:
            obj2 = s3.copy_object(Bucket=dest_bucket, Key=dest_key, CopySource=source_dict, MetadataDirective='COPY')
            break
        except bc_exceptions.ConnectionClosedError as err:
            print(err)
            counter = counter - 1
            if counter == 0:
                raise err
            print('...trying again...')
            sleep(5)
        except bc_exceptions.ConnectTimeoutError as err:
            print(err)
            counter = counter - 1
            if counter == 0:
                raise err
            print('...trying again...')
            sleep(5)

    return obj2


def put_interim_results_s3(s3, bucket, file_path, run_id, system_version=4, retries=5):
    """

    """
    path1 = pathlib.Path(file_path)
    ds_id, stn_id, start_date = path1.stem.split('.')[0].split('_')
    key = tdm.utils.key_patterns[system_version]['interim_results'].format(run_id=run_id, dataset_id=ds_id, station_id=stn_id, start_date=start_date)

    stats = os.stat(file_path)
    run_date = pd.Timestamp(round(stats.st_mtime), unit='s')
    run_date_key = make_run_date_key(run_date)

    _ = put_file_s3(s3, bucket, key, file_path, {'run_date': run_date_key}, 'application/zstd', retries=retries)

    return key


def list_objects_s3(s3, bucket, prefix, start_after='', delimiter='', continuation_token='', date_format=None):
    """
    Wrapper S3 function around the list_objects_v2 base function with a Pandas DataFrame output.

    Parameters
    ----------
    s3_client : boto3.client
        A boto3 client object
    bucket : str
        The S3 bucket.
    prefix : str
        Limits the response to keys that begin with the specified prefix.
    start_after : str
        The S3 key to start after.
    delimiter : str
        A delimiter is a character you use to group keys.
    continuation_token : str
        ContinuationToken indicates to S3 that the list is being continued on this bucket with a token.

    Returns
    -------
    DataFrame
    """
    if s3._endpoint.host == 'https://vault.revera.co.nz':
        js = []
        while True:
            js1 = s3.list_objects(Bucket=bucket, Prefix=prefix, Marker=start_after, Delimiter=delimiter)

            if 'Contents' in js1:
                js.extend(js1['Contents'])
                if 'NextMarker' in js1:
                    start_after = js1['NextMarker']
                else:
                    break
            else:
                break

    else:
        js = []
        while True:
            js1 = s3.list_objects_v2(Bucket=bucket, Prefix=prefix, StartAfter=start_after, Delimiter=delimiter, ContinuationToken=continuation_token)

            if 'Contents' in js1:
                js.extend(js1['Contents'])
                if 'NextContinuationToken' in js1:
                    continuation_token = js1['NextContinuationToken']
                else:
                    break
            else:
                break

    if js:
        f_df1 = pd.DataFrame(js)[['Key', 'LastModified', 'ETag', 'Size']].copy()
        if isinstance(date_format, str):
            f_df1['KeyDate'] = pd.to_datetime(f_df1.Key.apply(lambda x: path_date_parser(x, date_format)), utc=True, errors='coerce').dt.tz_localize(None)
        f_df1['ETag'] = f_df1['ETag'].str.replace('"', '')
        f_df1['LastModified'] = pd.to_datetime(f_df1['LastModified']).dt.tz_localize(None)
    else:
        if isinstance(date_format, str):
            f_df1 = pd.DataFrame(columns=['Key', 'LastModified', 'ETag', 'Size', 'KeyDate'])
        else:
            f_df1 = pd.DataFrame(columns=['Key', 'LastModified', 'ETag', 'Size'])

    return f_df1


def list_object_versions_s3(s3_client, bucket, prefix, start_after='', delimiter=None, date_format=None):
    """
    Wrapper S3 function around the list_object_versions base function with a Pandas DataFrame output.

    Parameters
    ----------
    s3_client : boto3.client
        A boto3 client object
    bucket : str
        The S3 bucket.
    prefix : str
        Limits the response to keys that begin with the specified prefix.
    start_after : str
        The S3 key to start at.
    delimiter : str or None
        A delimiter is a character you use to group keys.

    Returns
    -------
    DataFrame
    """
    js = []
    while True:
        if isinstance(delimiter, str):
            js1 = s3_client.list_object_versions(Bucket=bucket, Prefix=prefix, KeyMarker=start_after, Delimiter=delimiter)
        else:
            js1 = s3_client.list_object_versions(Bucket=bucket, Prefix=prefix, KeyMarker=start_after)

        if 'Versions' in js1:
            js.extend(js1['Versions'])
            if 'NextKeyMarker' in js1:
                start_after = js1['NextKeyMarker']
            else:
                break
        else:
            break

    if js:
        f_df1 = pd.DataFrame(js)[['Key', 'VersionId', 'IsLatest', 'LastModified', 'ETag', 'Size']].copy()
        if isinstance(date_format, str):
            f_df1['KeyDate'] = pd.to_datetime(f_df1.Key.apply(lambda x: path_date_parser(x, date_format)), utc=True, errors='coerce').dt.tz_localize(None)
        f_df1['ETag'] = f_df1['ETag'].str.replace('"', '')
        f_df1['LastModified'] = pd.to_datetime(f_df1['LastModified']).dt.tz_localize(None)
    else:
        if isinstance(date_format, str):
            f_df1 = pd.DataFrame(columns=['Key', 'VersionId', 'IsLatest', 'LastModified', 'ETag', 'Size', 'KeyDate'])
        else:
            f_df1 = pd.DataFrame(columns=['Key', 'VersionId', 'IsLatest', 'LastModified', 'ETag', 'Size'])

    return f_df1


def put_dataset(dataset, bucket, s3=None, connection_config=None, public_url=None, system_version=4):
    """

    """
    run_date_key = make_run_date_key()

    dataset_id = dataset['dataset_id']

    ## Get the stations agg
    stns_key = tdm.utils.key_patterns[system_version]['stations'].format(dataset_id=dataset_id)

    obj1 = get_object_s3(stns_key, bucket, s3, connection_config, public_url)
    stns = read_json_zstd(obj1)

    ## generate stats for the dataset metadata
    ds_stats = stats_for_dataset_metadata(stns)

    if dataset['result_type'] == 'time_series':
        if 'spatial_resolution' in ds_stats:
            ds_stats.pop('spatial_resolution')

    dataset.update(ds_stats)

    ## Add version number
    dataset.update({'system_version': system_version})

    ## Check and create dataset metadata
    ds4 = tdm.dataset.Dataset(**dataset)

    ds5 = orjson.loads(ds4.json(exclude_none=True))

    ## Write the object
    ds_obj = write_json_zstd(ds5)

    ds_key = tdm.utils.key_patterns[system_version]['dataset'].format(dataset_id=dataset_id)

    _ = put_object_s3(s3, bucket, ds_key, ds_obj, {'run_date': run_date_key}, 'application/json')

    return ds5


def get_remote_dataset(bucket, s3=None, connection_config=None, public_url=None, dataset_id=None, key=None, system_version=4):
    """

    """
    if isinstance(dataset_id, str):
        key = tdm.utils.key_patterns[system_version]['dataset'].format(dataset_id=dataset_id)
    elif not isinstance(key, str):
        raise ValueError('dataset_id must be passed or key must be passed.')

    try:
        obj1 = get_object_s3(key, bucket, s3, connection_config, public_url)
        rem_ds = read_json_zstd(obj1)
    except:
        rem_ds = None

    return rem_ds


def get_remote_station(bucket, s3=None, connection_config=None, public_url=None, dataset_id=None, station_id=None, key=None, system_version=4):
    """

    """
    if isinstance(dataset_id, str) and isinstance(station_id, str):
        key = tdm.utils.key_patterns[system_version]['station'].format(dataset_id=dataset_id, station_id=station_id)
    elif not isinstance(key, str):
        raise ValueError('dataset_id and station_id must be passed or key must be passed.')

    try:
        obj1 = get_object_s3(key, bucket, s3, connection_config, public_url)
        rem_stn = read_json_zstd(obj1)
    except:
        rem_stn = None

    return rem_stn


def get_remote_results_chunks(bucket, s3=None, connection_config=None, public_url=None, dataset_id=None, key=None, system_version=4):
    """

    """
    if isinstance(dataset_id, str):
        key = tdm.utils.key_patterns[system_version]['results_chunks'].format(dataset_id=dataset_id)
    elif not isinstance(key, str):
        raise ValueError('dataset_id must be passed or key must be passed.')

    try:
        obj1 = get_object_s3(key, bucket, s3, connection_config, public_url)
        rem_ds = read_json_zstd(obj1)
    except:
        rem_ds = None

    return rem_ds



# def put_remote_station(s3, bucket, station, run_date=None, system_version=4):
#     """

#     """
#     run_date_key = make_run_date_key(run_date)

#     dataset_id = station['dataset_id']
#     station_id = station['station_id']

#     stn4 = Station(**station)

#     stn5 = orjson.loads(stn4.json(exclude_none=True))

#     stn_obj = write_json_zstd(stn5)

#     stn_key = tdm.utils.key_patterns[system_version]['station'].format(dataset_id=dataset_id, station_id=station_id)

#     _ = put_object_s3(s3, bucket, stn_key, stn_obj, {'run_date': run_date_key}, 'application/json')

#     return stn5


def put_remote_agg_stations(dataset_id, bucket, s3, public_url=None, threads=60, system_version=4):
    """

    """
    stn_key = tdm.utils.key_patterns[system_version]['station']
    stn_prefix = stn_key.split('{station_id}')[0].format(dataset_id=dataset_id)

    if system_version == 2:
        list1 = list_objects_s3(s3, bucket, stn_prefix)
    elif system_version >= 3:
        list1 = list_objects_s3(s3, bucket, stn_prefix, delimiter='/')
    else:
        raise ValueError('Wrong object structure version.')

    list2 = list1[list1.Key.str.contains('station.json.zst')].copy()

    if not list2.empty:

        ## Get all of the result chunks from the individual stations
        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
            futures = []

            for key in list2['Key'].tolist():
                f = executor.submit(get_remote_station, s3=s3, bucket=bucket, key=key, public_url=public_url, system_version=system_version)
                futures.append(f)

            runs = concurrent.futures.wait(futures)

        stn_list = [r.result() for r in runs[0]]

        ## Remove the results chunks from the stations
        _ = [s.pop('results_chunks') for s in stn_list]

        ## Save files
        # Stations
        stns_obj = write_json_zstd(stn_list)

        run_date_key = make_run_date_key()

        agg_stn_key = tdm.utils.key_patterns[system_version]['stations']

        stns_key = agg_stn_key.format(dataset_id=dataset_id)

        _ = put_object_s3(s3, bucket, stns_key, stns_obj, {'run_date': run_date_key}, 'application/json')

        return stn_list
    else:
        print('There are no stations files in the database.')
        return None


def put_remote_agg_datasets(bucket, s3, public_url=None, threads=30, system_version=4):
    """

    """
    ds_key = tdm.utils.key_patterns[system_version]['dataset']

    ds_prefix = ds_key.split('{dataset_id}')[0]

    if system_version == 2:
        list1 = list_objects_s3(s3, bucket, ds_prefix)
    elif system_version >= 3:
        list1 = list_objects_s3(s3, bucket, ds_prefix, delimiter='/')
    else:
        raise ValueError('Wrong object structure version.')

    list2 = list1[list1.Key.str.contains('dataset.json.zst')].copy()

    if not list2.empty:

        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
            futures = []

            for key in list2['Key'].tolist():
                f = executor.submit(get_remote_dataset, s3=s3, bucket=bucket, public_url=public_url, key=key, system_version=system_version)
                futures.append(f)

            runs = concurrent.futures.wait(futures)

        ds_list = [r.result() for r in runs[0]]

        dss_obj = write_json_zstd(ds_list)

        run_date_key = make_run_date_key()
        dss_key = tdm.utils.key_patterns[system_version]['datasets']

        _ = put_object_s3(s3, bucket, dss_key, dss_obj, {'run_date': run_date_key}, 'application/json')

    return ds_list


def compare_datasets_from_s3(connection_config, bucket, new_data, add_old=False, last_run_date_key=None, public_url=None, version=3):
    """
    Parameters
    ----------
    connection_config : dict
        A dictionary of the connection info necessary to establish an S3 connection.
    bucket : str
        The S3 bucket.
    new_data : xr.Dataset
        The new data that should be compared to existing data in S3.
    add_old : bool
        Should the data in the S3 be added to the output?
    last_run_date_key : str
        Specify the last run key instead of having the function figure it out. The function will do a check to make sure that the key exists.
    public_url : str
        Optional if there is a public URL to the object instead of using the S3 API directly.

    Returns
    -------
    xr.Dataset
        Of the data that should be updated.
    """
    ## Determine the parameter, station_id, and dataset_id
    vars1 = list(new_data.variables)
    dataset = [new_data[v].attrs for v in vars1 if 'dataset_id' in new_data[v].attrs][0]
    dataset_id = dataset['dataset_id']
    # result_type = dataset['result_type']
    station_id = str(new_data.squeeze()['station_id'].values)

    key_dict = {'dataset_id': dataset_id, 'station_id': station_id}

    base_key_pattern = tdm.utils.key_patterns[version]['results']

    ## Get list of keys
    if isinstance(last_run_date_key, str):
        key_dict.update({'run_date': last_run_date_key})
        last_key = base_key_pattern.format(**key_dict)
    else:
        last_key = None

    ## Get previous data and compare
    if isinstance(public_url, str):
        conn_config = public_url
    else:
        conn_config = connection_config

    if isinstance(last_key, str):
        # last_key = last_key1.iloc[0]['Key']
        p_old_one = get_object_s3(last_key, conn_config, bucket, 'zstd')
        xr_old_one = xr.load_dataset(p_old_one)

        up1 = compare_xrs(xr_old_one, new_data, add_old=add_old)
    else:
        # print('No prior data found in S3. All data will be returned.')
        up1 = new_data.copy()

    return up1


# def update_results_s3(processing_code, data_dict, run_date_dict, remote, threads=30, public_url=None, version=3):
#     """
#     Parameters
#     ----------
#     processing_code : int
#         The processing code to determine how the input data should be processed.
#     data_dict : dict of lists
#         A dictionary with the keys as the dataset_ids and teh values as lists of zstd compressed xr.Datasets.
#     remote : dict
#         Dict of a connection_config and bucket:
#         conn_config : dict
#             A dictionary of the connection info necessary to establish an S3 connection.
#         bucket : str
#             The S3 bucket.
#     threads : int
#         The number of threads to use to process the data.
#     public_url : str
#         Optional if there is a public URL to the object instead of using the S3 API directly.

#     Returns
#     -------
#     None

#     """
#     ### Parameters
#     conn_config = remote['connection_config']
#     bucket = remote['bucket']

#     if processing_code in [2, 3, 6]:
#         add_old = True
#     elif processing_code in [1, 4, 5]:
#         add_old = False
#     else:
#         raise ValueError('processing_code does not exist.')

#     if processing_code in [4, 5]:
#         no_comparison = True
#     else:
#         no_comparison = False

#     ### Run update

#     s3 = s3_client(conn_config, threads)

#     for ds_id, results in data_dict.items():
#         print('--dataset_id: ' + ds_id)

#         if len(results) > 0:

#             run_date_key = run_date_dict[ds_id]

#             # Create the Key info dict
#             prefix = tdm.utils.key_patterns[version]['results'].split('{station_id}')[0].format(dataset_id=ds_id)

#             keys1 = list_objects_s3(s3, bucket, prefix)
#             obj_df1 = keys1[keys1.Key.str.contains('results.nc')].copy()
#             # obj_df1['dataset_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[2])
#             obj_df1['station_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[3])

#             last_date1 = obj_df1.groupby(['station_id'])['KeyDate'].last().reset_index()


#             def update_result(result):
#                 """

#                 """
#                 ## Process data
#                 try:
#                     new1 = xr.load_dataset(read_pkl_zstd(result, False))
#                 except:
#                     print('Data could not be opened')
#                     return None

#                 stn_id = str(new1.squeeze()['station_id'].values)
#                 # print('station_id: ' + stn_id)

#                 vars1 = list(new1.variables)
#                 parameter = [v for v in vars1 if 'dataset_id' in new1[v].attrs][0]
#                 result_attrs = new1[parameter].attrs.copy()

#                 ds_id = result_attrs['dataset_id']
#                 mod_date_key = new1.attrs['history'].split(':')[0]

#                 if no_comparison:
#                     up1 = new1
#                 else:
#                     last_date_key_df = last_date1[last_date1['station_id'] == stn_id]
#                     if last_date_key_df.empty:
#                         last_date_key = None
#                     else:
#                         last_date_key = make_run_date_key(last_date_key_df['KeyDate'].iloc[0])

#                     up1 = compare_datasets_from_s3(conn_config, bucket, new1, add_old=add_old, last_run_date_key=last_date_key, public_url=public_url, version=version)

#                 ## Save results
#                 if isinstance(up1, xr.Dataset) and (len(up1[parameter].time) > 0):

#                     # print('Save results')
#                     key_dict = {'dataset_id': ds_id, 'station_id': stn_id, 'run_date': run_date_key}

#                     new_key = tdm.utils.key_patterns[version]['results'].format(**key_dict)

#                     cctx = zstd.ZstdCompressor(level=1)
#                     c_obj = cctx.compress(up1.to_netcdf())

#                     obj_resp = put_object_s3(s3, bucket, new_key, c_obj, {'run_date': mod_date_key}, 'application/zstd')

#                     ## Process stn data
#                     # print('Save station data')

#                     ## Process object key infos
#                     stn_obj_df1 = obj_df1[obj_df1['station_id'] == stn_id].drop(['LastModified', 'station_id'], axis=1).copy()
#                     keys = stn_obj_df1['Key'].unique()
#                     new_key_info = [new_key, obj_resp['ResponseMetadata']['HTTPHeaders']['etag'].replace('"', ''), len(c_obj), pd.Timestamp(run_date_key).tz_localize(None)]

#                     if new_key in keys:
#                         stn_obj_df1[stn_obj_df1['Key'] == new_key] = new_key_info
#                     else:
#                         stn_obj_df1.loc[len(stn_obj_df1)] = new_key_info

#                     info1 = [S3ObjectKey(key=row['Key'], bucket=bucket, content_length=row['Size'], etag=row['ETag'], run_date=row['KeyDate']) for i, row in stn_obj_df1.iterrows()]

#                     ## Final station processing and saving
#                     stn_m = process_station_summ(up1, info1, mod_date=mod_date_key)

#                     up_stns = put_remote_station(s3, bucket, stn_m, run_date=mod_date_key, version=version)

#                 else:
#                     print('No new data to update')

#                 ## Get rid of big objects
#                 new1 = None
#                 up1 = None


#             ## Run the threadpool
#             # output = ThreadPool(threads).imap_unordered(update_result, results)
#             with ThreadPool(threads) as pool:
#                 output = pool.map(update_result, results)
#                 pool.close()
#                 pool.join()


def update_versions(version_data, bucket, connection_config, public_url=None, system_version=4):
    """

    """
    ## Check version_dict data model
    version_dict1 = tdm.dataset.ResultVersion(**version_data).dict(exclude_none=True)

    ## Initial set up
    s3 = s3_client(connection_config)

    dataset_id = version_dict1['dataset_id']

    ## Get the versions
    rv_key = tdm.utils.key_patterns[system_version]['versions'].format(dataset_id=dataset_id)

    v_obj1 = get_object_s3(rv_key, bucket, s3=s3, public_url=public_url, counter=1)

    if v_obj1 is not None:
        v2 = read_json_zstd(v_obj1)

        version_list = []
        for v in v2:
            if version_dict1['version_date'] == v['version_date']:
                version_list.append(version_dict1)
            else:
                version_list.append(v)
    else:
        version_list = [version_dict1]

    ## Save files
    run_date_key = make_run_date_key()

    # versions
    rv_obj = write_json_zstd(version_list)

    _ = put_object_s3(s3, bucket, rv_key, rv_obj, {'run_date': run_date_key}, 'application/json')

    return version_list


def put_result(s3, bucket, results_path, system_version=4):
    """

    """
    file_name = os.path.split(results_path)[1]
    ds_id, version_date_key, stn_id, chunk_id, hash1, _ = file_name.split('_')

    key_name = tdm.utils.key_patterns[system_version]['results'].format(dataset_id=ds_id, version_date=version_date_key, station_id=stn_id, chunk_id=chunk_id)

    metadata = {'chunk_hash': hash1}

    _ = put_file_s3(s3, bucket, key_name, results_path, metadata, 'application/zstd', retries=5)

    return key_name


def put_station(s3, bucket, stn_path, system_version=4):
    """

    """
    file_name = os.path.split(stn_path)[1]
    ds_id, stn_id, _ = file_name.split('_')

    key_name = tdm.utils.key_patterns[system_version]['station'].format(dataset_id=ds_id, station_id=stn_id)

    metadata = {}

    _ = put_file_s3(s3, bucket, key_name, stn_path, metadata, 'application/zstd', retries=5)

    return key_name


def delete_result_objects_s3(conn_config, bucket, dataset_ids=None, keep_last=10, threads=50, version=3):
    """
    Function to delete Tethys result objects including all object versions.

    Parameters
    ----------
    conn_config : dict
        A dictionary of the connection info necessary to establish an S3 connection.
    bucket : str
        The s3 bucket.
    dataset_ids : list, str, or None
        The specific datasets that should have the results objects removed. None will remove results objects from all datasets in the bucket.
    keep_last : int
        That last number of runs that should be kept. E.g. a value of 4 will kepp the last 4 runs and remove all prior runs.
    threads : int
        The number of concurrent threads to use when aggregating the stations and datasets.

    Returns
    -------
    list of keys deleted
    """
    s3 = s3_client(conn_config, threads)

    if isinstance(dataset_ids, str):
        dataset_ids = [dataset_ids]

    prefix = tdm.utils.key_patterns[version]['results'].split('{dataset_id}')[0]

    obj_list = list_object_versions_s3(s3, bucket, prefix)

    obj_list1 = obj_list[obj_list.KeyDate.notnull()].copy()
    key_split = obj_list1['Key'].str.split('/')
    obj_list1['dataset_id'] = key_split.apply(lambda x: x[2])
    obj_list1['station_id'] = key_split.apply(lambda x: x[3])

    if isinstance(dataset_ids, list):
        obj_list1 = obj_list1[obj_list1['dataset_id'].isin(dataset_ids)].copy()

    ## Get the keys to the objects that should be removed
    obj_list2 = obj_list1.groupby(['dataset_id', 'station_id'])

    rem_keys = []
    for i, row in obj_list2:
        rem1 = row.sort_values('KeyDate', ascending=False).iloc[keep_last:]
        for i2, row2 in rem1.iterrows():
            rem_keys.extend([{'Key': row2['Key'], 'VersionId': row2['VersionId']}])

    if len(rem_keys) > 0:
        ## Split them into 1000 key chunks
        rem_keys_chunks = np.array_split(rem_keys, int(np.ceil(len(rem_keys)/1000)))

        ## Run through and delete the objects...
        for keys in rem_keys_chunks:
            resp = s3.delete_objects(Bucket=bucket, Delete={'Objects': keys.tolist(), 'Quiet': True})

    print(str(len(rem_keys)) + ' objects removed')
    print('Updating stations and dataset json files...')

    dataset_list = obj_list1['dataset_id'].unique().tolist()

    for ds in dataset_list:
        ds_stations = put_remote_agg_stations(s3, bucket, ds, threads)

    ### Aggregate all datasets for the bucket
    ds_all = put_remote_agg_datasets(s3, bucket, threads)

    return rem_keys


# def process_results_versions(dataset_list, connection_config, bucket, version=4, version_date=None, version_data=None):
#     """
#     Function to process the run date keys for all datasets for the extraction. These are specific to each processing_code.

#     Parameters
#     ----------
#     processing_code : int
#         The processing code to determine how the input data should be processed.
#     dataset_list : list of dict
#         The list of datasets, which is the output of the process_datasets function.
#     remote : dict
#         Dict of a connection_config and bucket:
#         conn_config : dict
#             A dictionary of the connection info necessary to establish an S3 connection.
#         bucket : str
#             The S3 bucket.
#     run_date : str, datetime, date, pd.Timestamp, or None
#         The run_date to use for processing. If None, then the run_date will be generated.
#     save_interval_hours : int
#         The frequency by which the datasets should have run dates saved as in Tethys. Essentially the question is, how frequently do you want to keep a record of the data changes? This value is in hours and the default is 2 weeks. This does not effect datasets with processing_code 3 as the run_date always stays the same. A processing_code of 4 oe 5 always creates a new run_date.

#     Returns
#     -------
#     run_date_dict : dict of str
#     """
#     if isinstance(version_date, (str, datetime, date, pd.Timestamp)):
#         run_date1 = pd.Timestamp(version_date)
#     else:
#         run_date1 = pd.Timestamp.today(tz='utc').tz_localize(None).round('s')

#     run_date_dict = {}
#     for ds in dataset_list:
#         dataset_id = ds['dataset_id']

#         ## Get last result dates
#         try:
#             key1 = tdm.utils.key_patterns[version]['stations'].format(dataset_id=dataset_id)
#             obj1 = get_object_s3(key1, connection_config, bucket, counter=0)
#             rok1 = read_json_zstd(obj1)
#             if version <= 3:
#                 rok1b = [r['results_object_key'] if isinstance(r['results_object_key'], dict) else r['results_object_key'][-1] for r in rok1]
#                 rok2 = pd.DataFrame(rok1b)
#                 rok2['run_date'] = pd.to_datetime(rok2['run_date'])
#                 obj_list = rok2.dropna(subset=['run_date'])
#                 last_run_date = obj_list['run_date'].max()
#             else:
#                 rok1b = [r['result_versions'] for r in rok1]
#                 rok2 = pd.DataFrame(rok1b)
#                 rok2['version_date'] = pd.to_datetime(rok2['version_date'])
#                 obj_list = rok2.dropna(subset=['version_date'])
#                 last_run_date = obj_list['version_date'].max()
#         except:
#             last_run_date = run_date1

#         if processing_code in [3]:
#             last_run_date_key = last_run_date.strftime('%Y%m%dT%H%M%SZ')
#         elif processing_code in [4, 5]:
#             last_run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
#         elif processing_code in [1, 2, 6]:
#             if last_run_date < (run_date1 - pd.DateOffset(hours=save_interval_hours) + pd.DateOffset(minutes=5)):
#                 last_run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
#             else:
#                 last_run_date_key = last_run_date.strftime('%Y%m%dT%H%M%SZ')
#         else:
#             raise ValueError('processing_code does not exist.')

#         run_date_dict.update({dataset_id: last_run_date_key})

#     return run_date_dict


def process_results_versions(dataset_list, bucket, s3=None, connection_config=None, public_url=None, version_data=None, system_version=4):
    """
    Function to process the run date keys for all datasets for the extraction. These are specific to each processing_code.

    Parameters
    ----------
    processing_code : int
        The processing code to determine how the input data should be processed.
    dataset_list : list of dict
        The list of datasets, which is the output of the process_datasets function.
    remote : dict
        Dict of a connection_config and bucket:
        conn_config : dict
            A dictionary of the connection info necessary to establish an S3 connection.
        bucket : str
            The S3 bucket.
    run_date : str, datetime, date, pd.Timestamp, or None
        The run_date to use for processing. If None, then the run_date will be generated.


    Returns
    -------
    run_date_dict : dict of str
    """
    mod_date = pd.Timestamp.today(tz='utc').tz_localize(None).round('s')

    if isinstance(version_data, dict):
        if isinstance(version_data['version_date'], str):
            version_data['version_date'] = pd.Timestamp(version_data['version_date'])

        version_data['modified_date'] = mod_date

        version_dict = {}
        for ds in dataset_list:
            dataset_id = ds['dataset_id']
            vd = version_data.copy()
            vd['dataset_id'] = dataset_id
            version_m = tdm.dataset.ResultVersion(**vd)
            version_dict[dataset_id] = orjson.loads(version_m.json(exclude_none=True))

    else:
        version_dict = {'modified_date': mod_date}
        for ds in dataset_list:
            dataset_id = ds['dataset_id']
            vd = {'dataset_id': dataset_id}

            ## Get last result dates
            ## I probably need to make variants for versions 2 and 3
            if system_version == 4:
                try:
                    key1 = tdm.utils.key_patterns[system_version]['results_versions'].format(dataset_id=dataset_id)
                    obj1 = get_object_s3(key1, bucket, s3, connection_config, public_url, counter=1)
                    rok1 = read_json_zstd(obj1)

                    version_date = rok1[-1]['version_date']
                    vd['version_date'] = version_date
                    version_m = tdm.dataset.ResultVersion(**vd)
                    version_dict[dataset_id] = orjson.loads(version_m.json(exclude_none=True))
                except:
                    run_date1 = pd.Timestamp.today(tz='utc').tz_localize(None).round('s')
                    vd['version_date'] = run_date1
                    version_m = tdm.dataset.ResultVersion(**vd)
                    version_dict[dataset_id] = orjson.loads(version_m.json(exclude_none=True))
            else:
                raise NotImplementedError('I need to make variants for versions 2 and 3.')

    return version_dict


# def reprocess_datasets(dataset, conn_config, bucket, public_url, threads=20):
#     """

#     """
#     dataset_id = dataset['dataset_id']
#     tethys = Tethys([{'connection_config': public_url, 'bucket': bucket, 'version': 2}])

#     stns = tethys.get_stations(dataset_id, results_object_keys=True)

#     bad_stns = [s for s in stns if 'time_range' not in s]

#     prefix = tdm.utils.key_patterns[2]['results'].split('{station_id}')[0].format(dataset_id=dataset_id)

#     s3 = s3_client(conn_config)
#     keys1 = list_objects_s3(s3, bucket, prefix)
#     obj_df1 = keys1[keys1.Key.str.contains('results.nc')].copy()
#     obj_df1['station_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[3])


#     def reprocess_dataset(stn):
#         """

#         """
#         stn_id = stn['station_id']
#         run_dates = tethys.get_run_dates(dataset_id, stn['station_id'])

#         for run_date in run_dates:

#         # run_date = stn['results_object_key']['run_date']
#             run_date_key = make_run_date_key(run_date)
#             mod_date_key = make_run_date_key()
#             up1 = tethys.get_results(dataset_id, stn_id, run_date=run_date, output='Dataset')

#             key_dict = {'dataset_id': dataset_id, 'station_id': stn_id, 'run_date': run_date_key}

#             new_key = tdm.utils.key_patterns[2]['results'].format(**key_dict)

#             cctx = zstd.ZstdCompressor(level=1)
#             c_obj = cctx.compress(up1.to_netcdf())

#             obj_resp = put_object_s3(s3, bucket, new_key, c_obj, {'run_date': mod_date_key}, 'application/zstd')

#             ## Process stn data
#             # print('Save station data')

#         ## Process object key infos
#         stn_obj_df1 = obj_df1[obj_df1['station_id'] == stn_id].drop(['LastModified', 'station_id'], axis=1).copy()
#         keys = stn_obj_df1['Key'].unique()
#         new_key_info = [new_key, obj_resp['ResponseMetadata']['HTTPHeaders']['etag'].replace('"', ''), len(c_obj), pd.Timestamp(run_date_key).tz_localize(None)]

#         if new_key in keys:
#             stn_obj_df1[stn_obj_df1['Key'] == new_key] = new_key_info
#         else:
#             stn_obj_df1.loc[len(stn_obj_df1)] = new_key_info

#         info1 = [S3ObjectKey(key=row['Key'], bucket=bucket, content_length=row['Size'], etag=row['ETag'], run_date=row['KeyDate']) for i, row in stn_obj_df1.iterrows()]

#         ## Final station processing and saving
#         stn_m = process_station_summ(up1, info1, mod_date=mod_date_key)

#         up_stns = put_remote_station(s3, bucket, stn_m, run_date=mod_date_key, version=2)


#     if len(bad_stns) > 0:
#         print('Updating results and stations...')
#         output = ThreadPool(threads).map(reprocess_dataset, bad_stns)

#         print('Updating aggregates...')

#         ds_stations = put_remote_agg_stations(s3, bucket, dataset_id, threads*2, version=2)
#         if ds_stations is not None:
#             ds_new = put_remote_dataset(s3, bucket, dataset, ds_stations, version=2)

#         # Aggregate all datasets for the bucket
#         ds_all = put_remote_agg_datasets(s3, bucket, threads*2, version=2)

#     else:
#         print('Nothing to update.')

#     print('Finished')


def determine_results_chunks_diffs(source_paths, remote):
    """

    """
    paths1 = []
    for p in source_paths:
        ds_id, version_date1, stn_id, chunk_id, hash_id, _ = os.path.split(p)[1].split('_')
        paths1.append([p, ds_id, version_date1, stn_id, chunk_id, hash_id])

    paths2 = pd.DataFrame(paths1, columns=['file_path', 'dataset_id', 'version_date', 'station_id', 'chunk_id', 'chunk_hash'])
    paths2['version_date'] = pd.to_datetime(paths2['version_date']).dt.tz_localize(None)

    tethys = Tethys([remote])
    datasets = tethys._datasets

    diffs_list = []
    for ds_id, g in paths2.groupby('dataset_id'):
        # print(ds_id)
        if ds_id in datasets:
            _ = tethys.get_versions(ds_id)
            vd_max = g.version_date.max()

            chunks1 = []
            for stn_id, c in tethys._results_chunks[ds_id].items():
                # print(stn_id)
                chunks2 = chunk_filters(c, version_date=vd_max)
                chunks1.extend(chunks2)

            chunks_df = pd.DataFrame(chunks1).drop(['version_date', 'content_length', 'height', 'chunk_day'], axis=1).rename(columns={'chunk_hash': 'original_chunk_hash'})

            combo1 = pd.merge(g, chunks_df, on=['dataset_id', 'station_id', 'chunk_id'], indicator=True, how='outer')
            combo1 = combo1[combo1['_merge'].isin(['both', 'left_only'])]
            combo1 = combo1[combo1['chunk_hash'] != combo1['original_chunk_hash']].copy()

            diffs_list.append(combo1.drop(['key', 'original_chunk_hash', '_merge'], axis=1))
        else:
            diffs_list.append(g)

    combo2 = pd.concat(diffs_list)

    return combo2








# for ds in titan.dataset_list:
#     reprocess_datasets(ds['dataset_id'], conn_config, bucket, public_url, threads=20)





# for result in results[2500:]:
#     new1 = xr.load_dataset(read_pkl_zstd(result, False))
#
#     stn_id = str(new1.squeeze()['station_id'].values)
#     print('station_id: ' + stn_id)
#
#     vars1 = list(new1.variables)
#     parameter = [v for v in vars1 if 'dataset_id' in new1[v].attrs][0]
#     result_attrs = new1[parameter].attrs.copy()
#
#     ds_id = result_attrs['dataset_id']
#     mod_date_key = new1.attrs['history'].split(':')[0]
#
#     if no_comparison:
#         up1 = new1
#     else:
#         last_date_key_df = last_date1[last_date1['station_id'] == stn_id]
#         if last_date_key_df.empty:
#             last_date_key = None
#         else:
#             last_date_key = make_run_date_key(last_date_key_df['KeyDate'].iloc[0])
#
#         up1 = compare_datasets_from_s3(conn_config, bucket, new1, add_old=add_old, last_run_date_key=last_date_key, public_url=public_url)
#
#     ## Save results
#     if isinstance(up1, xr.Dataset) and (len(up1[parameter].time) > 0):
#
#         # print('Save results')
#         key_dict = {'dataset_id': ds_id, 'station_id': stn_id, 'run_date': run_date_key}
#
#         new_key = tdm.utils.key_patterns['results'].format(**key_dict)
#
#         cctx = zstd.ZstdCompressor(level=1)
#         c_obj = cctx.compress(up1.to_netcdf())
#
#         obj_resp = s3.put_object(Body=c_obj, Bucket=bucket, Key=new_key, ContentType='application/zstd', Metadata={'run_date': mod_date_key})
#
#         ## Process stn data
#         # print('Save station data')
#
#         ## Process object key infos
#         stn_obj_df1 = obj_df1[obj_df1['station_id'] == stn_id].drop(['LastModified', 'station_id'], axis=1).copy()
#         keys = stn_obj_df1['Key'].unique()
#         new_key_info = [new_key, obj_resp['ResponseMetadata']['HTTPHeaders']['etag'].replace('"', ''), len(c_obj), pd.Timestamp(run_date_key).tz_localize(None)]
#
#         if new_key in keys:
#             stn_obj_df1[stn_obj_df1['Key'] == new_key] = new_key_info
#         else:
#             stn_obj_df1.loc[len(stn_obj_df1)] = new_key_info
#
#         info1 = [S3ObjectKey(key=row['Key'], bucket=bucket, content_length=row['Size'], etag=row['ETag'], run_date=row['KeyDate']) for i, row in stn_obj_df1.iterrows()]
#
#         ## Final station processing and saving
#         stn_m = process_station_summ(ds_id, stn_id, up1, info1, mod_date=mod_date_key)
#
#         up_stns = put_remote_station(s3, bucket, stn_m, run_date=mod_date_key)
#
#     else:
#         print('No new data to update')
#
#     ## Get rid of big objects
#     new1 = None
#     up1 = None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  7 09:29:29 2021

@author: mike
"""
import os
import types
import numpy as np
import xarray as xr
import pandas as pd
from tethys_utils.titan import Titan
from tethys_utils.misc import parse_data_paths
# from shapely.geometry import shape, mapping, Point, box
import copy
# import rasterio
# import concurrent.futures
# from uuid import uuid4
# from dask.diagnostics import ProgressBar
import tethys_utils as tu
# from pathlib import Path
from glob import glob


###########################################
### Parameters

base_encoding = {'lon': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 1e-07},
'lat': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 1e-07},
'altitude': {'dtype': 'int32', '_FillValue': -9999, 'scale_factor': 0.001},
'time': {'_FillValue': -99999999, 'units': 'days since 1970-01-01 00:00:00'},
'height': {'dtype': 'int16', '_FillValue': -9999}}

############################################
### Functions


def cust_range(*args, rtol=1e-05, atol=1e-08, include=[True, True]):
    """
    Combines numpy.arange and numpy.isclose to mimic
    open, half-open and closed intervals.
    Avoids also floating point rounding errors as with
    >>> numpy.arange(1, 1.3, 0.1)
    array([1. , 1.1, 1.2, 1.3])

    args: [start, ]stop, [step, ]
        as in numpy.arange
    rtol, atol: floats
        floating point tolerance as in numpy.isclose
    include: boolean list-like, length 2
        if start and end point are included
    """
    # process arguments
    if len(args) == 1:
        start = 0
        stop = args[0]
        step = 1
    elif len(args) == 2:
        start, stop = args
        step = 1
    else:
        assert len(args) == 3
        start, stop, step = tuple(args)

    # determine number of segments
    n = (stop-start)/step + 1

    # do rounding for n
    if np.isclose(n, np.round(n), rtol=rtol, atol=atol):
        n = np.round(n)

    # correct for start/end is exluded
    if not include[0]:
        n -= 1
        start += step
    if not include[1]:
        n -= 1
        stop -= step

    return np.linspace(start, stop, int(n))


def split_grid(ds, x_size=None, y_size=None, n_intervals=None, x_name='lon', y_name='lat', mbytes_size=2000):
    """
    Function to split an n-dimensional dataset along the x and y dimensions. Optionally, add time and height dimensions if the array does not aready contain them.

    Parameters
    ----------
    ds : DataSet
        An xarray DataSet with at least x and y dimensions. It can have any number of dimensions, though it probably does not make much sense to have greater than 4 dimensions.
    x_size : int
        The size or length of the smaller grids in the x dimension.
    y_size : int
        The size or length of the smaller grids in the y dimension.
    x_name : str
        The x dimension name.
    y_name : str
        The y dimension name.

    Returns
    -------
    List of DataArrays
        The result contains none of the original attributes.
    """
    ## Get the dimension data
    # dims = ds.dims

    # Get other array info
    # x_index = dims[x_name]
    # y_index = dims[y_name]

    x_min = ds[x_name][0].values
    x_max = ds[x_name][-1].values
    y_min = ds[y_name][0].values
    y_max = ds[y_name][-1].values

    y_diff = ds[y_name].diff(y_name, 1).mean().round(5).values
    x_diff = ds[x_name].diff(x_name, 1).mean().round(5).values

    # Get the aranges of the dimensions
    if isinstance(x_size, int) and isinstance(y_size, int):
        x_range = cust_range(x_min, x_max+x_diff, x_diff*x_size) - x_diff/2
        y_range = cust_range(y_min, y_max+y_diff, y_diff*y_size) - y_diff/2

    else:
        if not isinstance(n_intervals, int):
            print('n_intervals will be calculated.')
            n_intervals = round(np.sqrt(ds.nbytes/(mbytes_size*1000000)))
        x_range = np.linspace(x_min, x_max, n_intervals+1, endpoint=True) - (x_diff/2)
        x_range[-1] = x_range[-1] + x_diff
        y_range = np.linspace(y_min, y_max, n_intervals+1, endpoint=True) - (y_diff/2)
        y_range[-1] = y_range[-1] + y_diff

    block_list = []
    for iy, y in enumerate(y_range[1:]):
        ds2 = ds.sel({y_name: slice(y_range[iy], y_range[iy+1])})
        for ix, x in enumerate(x_range[1:]):
            ds3 = ds2.sel({x_name: slice(x_range[ix], x_range[ix+1])})
            if not 0 in list(ds3.dims.values()):
                block_list.append(ds3)

    return block_list


def select_sample_points(grid, block_size=10):
    """

    """
    g1 = grid.isel(time=0, height=0, drop=True).copy().load()
    var1 = [v for v in list(g1.variables) if not v in ['lat', 'lon']][0]

    blocks = split_grid(g1, block_size, block_size)
    blocks1 = [b for b in blocks if int(b[var1].isnull().sum()) == 0]

    r1 = np.random.choice(range(len(blocks1)), 1)[0]

    block = blocks1[r1]

    lons = block['lon'].values
    lats = block['lat'].values

    # df1 = g1.to_dataframe().dropna().reset_index()
    # r1 = np.random.choice(range(len(df1)), samples, False)
    # df2 = df1.iloc[r1][['lon', 'lat']].reset_index(drop=True).copy()

    return lons, lats


def grid_checks(dims):
    """

    """
    if not 'lon' in dims:
        raise ValueError('lon must be a dimension.')
    if not 'lat' in dims:
        raise ValueError('lat must be a dimension.')
    if not 'time' in dims:
        raise ValueError("time must be a dimension.")
    if not 'height' in dims:
        raise ValueError("height must be a dimension.")


def iter_obj_sizes(lons_lats, grid, encoding):
    """

    """
    lons, lats = lons_lats

    block_size = len(lons)

    val1 = grid.sel(lon=lons, lat=lats).copy().load()

    for k, v in encoding.items():
            if k in val1:
                val1[k].encoding = v

    obj_sizes = {}
    for l in range(block_size):
        # print(l)
        lons1 = lons[:(l+1)]
        lats1 = lats[:(l+1)]

        val2 = val1.sel(lon=lons1, lat=lats1)
        obj_size1 = len(tu.processing.write_pkl_zstd(val2.to_netcdf()))

        obj_sizes.update({l+1: obj_size1})

    return obj_sizes




############################################
### Classes


class Processor(object):
    """

    """
    def __init__(self):
        """

        """
        # if isinstance(preprocessor, types.FunctionType):
        # self.preprocessor = preprocessor
        # if isinstance(postprocessor, types.FunctionType):
        # self.postprocessor = postprocessor
        # self.time = time
        # self.height = height

        pass


    def load_dataset_metadata(self, datasets):
        """

        """
        dataset_list = tu.processing.process_datasets(datasets)

        ds_dict = {ds['dataset_id']: ds for ds in dataset_list}

        setattr(self, 'dataset_list', dataset_list)
        setattr(self, 'datasets', ds_dict)

        ## Checks
        grid_bool = all([d['result_type'] == 'grid' for d in dataset_list])
        if not grid_bool:
            raise ValueError('All values of result_type in the datasets should be grid.')

        grouping_len = len(np.unique([d['grouping'] for d in dataset_list]))
        if grouping_len > 1:
            raise ValueError('All values of grouping in the datasets must be the same.')

        grouping = dataset_list[0]['grouping']

        # param_len = len(np.unique([d['parameter'] for d in dataset_list]))
        # if param_len > 1:
        #     raise ValueError('All values of parameter in the datasets must be the same.')

        setattr(self, 'grouping', grouping)

        return dataset_list


    def parse_paths(self, glob_obj, date_format=None, from_date=None, to_date=None):
        """

        """
        paths = parse_data_paths(glob_obj, date_format, from_date, to_date)

        self.source_data_paths = paths

        return paths


    def load_raw_grid(self, lon_dim_name, lat_dim_name, time_dim_name=None, chunks=None, parallel=True, preprocessor=None, postprocessor=None, **kwargs):
        """
        The processor takes a str path to one or many files that xr.open_mfdataset can open (e.g. netcdf) and runs a preprocessor and postprocessor (optionally) to provide the proper dimensions and names for the dataset. A Dataset can also be passed to data_path and this will skip the preprocessing step.
        """
        grid1 = xr.open_mfdataset(self.source_data_paths, chunks=chunks, parallel=parallel, preprocess=preprocessor)

        if isinstance(postprocessor, types.FunctionType):
            grid1 = postprocessor(grid1, **kwargs)

        ## add attributes
        self._dims = grid1.dims
        self._lon_dim_name = lon_dim_name
        self._lat_dim_name = lat_dim_name
        self._time_dim_name = time_dim_name
        self._preprocessor = preprocessor
        self._postprocessor = postprocessor
        self._kwargs = kwargs

        return grid1


    # def assess_input_grid(self, dataset_id, samples=900):
    #     """

    #     """
    #     ## checks
    #     if not dataset_id in self.datasets:
    #         raise ValueError('dataset_id not in datasets.')

    #     grid_checks(self._dims)

    #     ## some nice processing
    #     self._extract_results_distribution()


    def save_grid_chunks(self, output_path, block_size=None, block_size_multiplier=10):
        """

        """
        ## Checks
        grid_checks(self._dims)

        if not os.path.exists(output_path):
            os.makedirs(output_path)

        ## some nice processing
        if not hasattr(self, '_null_grid'):
            self._extract_null_grid()

        ## Parameters
        if block_size is None:
            obj_dict = self._obj_dict.copy()

            xy_chunk_size = int(obj_dict['block_size'] * block_size_multiplier)
        elif isinstance(block_size, int):
            xy_chunk_size = int(block_size * block_size_multiplier)

        chunks = {self._time_dim_name: self._dims['time'], self._lon_dim_name: xy_chunk_size, self._lat_dim_name: xy_chunk_size}

        ## Open grid
        grid1 = self.load_raw_grid(self._lon_dim_name, self._lat_dim_name, self._time_dim_name, chunks=chunks, parallel=True, preprocessor=self._preprocessor, postprocessor=self._postprocessor, **self._kwargs)

        ## Create the blocks
        block_list = split_grid(grid1, x_size=xy_chunk_size, y_size=xy_chunk_size, n_intervals=None, x_name='lon', y_name='lat', mbytes_size=1600)

        ## get null grid
        null_grid = self._null_grid

        ## Determine blocks that are not all null values
        null_list = split_grid(null_grid.to_dataset(), x_size=xy_chunk_size, y_size=xy_chunk_size, n_intervals=None, x_name='lon', y_name='lat', mbytes_size=1600)

        null_list1 = [bool(n1.null_grid.sum() != 0) for n1 in null_list]

        len1 = sum(null_list1)

        file_name_str = '{}_part_{:>0' + str(len(str(len1))) + 'd}.nc'

        print(str(len1) + ' grid block files will be saved.')

        block_list = [b for i, b in enumerate(block_list) if null_list1[i]]

        ## Determine which datasets are associated with the grid
        vars1 = [v for v in list(grid1.variables) if (not v in ('lon', 'lat')) and (len(grid1[v].dims) == 4)]
        dss = {ds['dataset_id']: list(ds['properties']['encoding'].keys())  for ds in self.dataset_list if ds['parameter'] in vars1}

        if len(dss) == 0:
            raise ValueError('No dataset metadata fit the dataset results.')

        for i, b in enumerate(block_list):
            c = b.copy()
            c.load()
            for ds_id, v in dss.items():
                file_name = file_name_str.format(ds_id, i+1)
                file_path = os.path.join(output_path, file_name)
                print(file_name)
                c[v].to_netcdf(file_path)
            c.close()
            c = None
            b.close()
            b = None

        grid1.close()
        grid1 = None


    def _select_sample_points(self, ds_vars, block_size=10):
        """

        """
        if isinstance(self._time_dim_name, str):
            chunks = {self._time_dim_name: 1}
        else:
            chunks = None

        grid1 = self.load_raw_grid(self._lon_dim_name, self._lat_dim_name, self._time_dim_name, chunks=chunks, parallel=True, preprocessor=self._preprocessor, postprocessor=self._postprocessor, **self._kwargs)
        grid1 = grid1[ds_vars]
        lons_lats = select_sample_points(grid1, block_size=block_size)

        self.sample_lons_lats = lons_lats

        grid1.close()
        grid1 = None

        return lons_lats


    def _extract_null_grid(self):
        """

        """
        if isinstance(self._time_dim_name, str):
            chunks = {self._time_dim_name: 1}
        else:
            chunks = None

        grid1 = self.load_raw_grid(self._lon_dim_name, self._lat_dim_name, self._time_dim_name, chunks=chunks, parallel=True, preprocessor=self._preprocessor, postprocessor=self._postprocessor, **self._kwargs)

        grid2 = grid1.isel(time=0, height=0, drop=True).copy()

        vars1 = [v for v in list(grid2.variables) if (not v in ('lon', 'lat')) and (len(grid2[v].dims) == 2)]
        grid3 = grid2[vars1[0]].load().notnull()
        grid3.name = 'null_grid'

        self._null_grid = grid3


    def determine_grid_block_size(self, min_size=1000, max_size=2000, block_size=10):
        """

        """
        ## checks
        # if not dataset_id in self.datasets:
        #     raise ValueError('dataset_id not in datasets.')

        grid_checks(self._dims)

        ## some nice processing
        self._extract_null_grid()

        ## iterate datasets
        ds_obj_sizes = {}

        dataset_list = copy.deepcopy(self.dataset_list)
        for dataset in dataset_list:
            dataset_id = dataset['dataset_id']
            print(dataset_id)

            dataset = self.datasets[dataset_id]
            dataset_enc = dataset['properties']['encoding']
            encoding = copy.deepcopy(base_encoding)
            encoding.update(dataset_enc)

            ds_vars = [list(ds['properties']['encoding'].keys())  for ds in self.dataset_list if ds['dataset_id'] == dataset_id][0]

            ## get sample points
            lons_lats = self._select_sample_points(ds_vars, block_size)

            # calc obj sizes
            grid1 = self.load_raw_grid(self._lon_dim_name, self._lat_dim_name, self._time_dim_name, chunks={self._time_dim_name: self._dims['time'], self._lon_dim_name: 1, self._lat_dim_name: 1}, parallel=True, preprocessor=self._preprocessor, postprocessor=self._postprocessor, **self._kwargs)

            grid2 = grid1[ds_vars]

            # Do a round of 4 for initial assessment
            obj_sizes = iter_obj_sizes(lons_lats, grid2, encoding)

            null_grid = self._null_grid
            n_stns = int(null_grid.sum())

            if (obj_sizes[1] > max_size*1000):
                block_size = 1
                mean_size = obj_sizes[1]
                obj_size = obj_sizes[1]

                [ds.update({'grouping': 'none'}) for ds in dataset_list]

                print('The results for this dataset should not be grouped and has been changed internally.')

            elif (obj_sizes[10] < min_size*1000):
                mean_size = obj_sizes[10]/100
                block_size = int(((max_size*1000)//mean_size)**0.5)
                obj_size = mean_size * block_size**2

            else:

                obj_sizes1 = {i: o for i, o in obj_sizes.items() if (o >  min_size*1000) and (o <  max_size*1000)}
                if len(obj_sizes1) == 0:
                    raise ValueError('No results for the block size. Widen the range of the min and max sizes.')

                block_size = max(obj_sizes1.keys())
                mean_size = obj_sizes[block_size]/(block_size**2)
                obj_size = obj_sizes[block_size]

                [ds.update({'grouping': 'blocks'}) for ds in dataset_list]

            print('block size estimate is ' + str(block_size))

            obj_dict = {'block_size': block_size, 'obj_size': obj_size, 'sum_obj_sizes': int(mean_size * n_stns), 'n_stations': n_stns}
            self._obj_dict = obj_dict

            _ = self.load_dataset_metadata(dataset_list)

            grid1.close()
            grid1 = None
            grid2.close()
            grid2 = None

            ds_obj_sizes.update({dataset_id: obj_dict})

        return ds_obj_sizes


class Grid(Titan):
    """

    """

    ## Initial import and assignment function
    # def __init__(self):
    #     """

    #     """
    #     # self._load_dataset_metadata = self.load_dataset_metadata
    #     # self._load_results = self.load_results

    #     # del self.load_dataset_metadata
    #     # delattr(Titan, 'load_dataset_metadata')
    #     # delattr(Titan, 'load_results')
    #     # if isinstance(preprocessor, types.FunctionType):
    #     # self.preprocessor = preprocessor
    #     # if isinstance(postprocessor, types.FunctionType):
    #     # self.postprocessor = postprocessor
    #     # self.time = time
    #     # self.height = height

    #     ## Assign grid resolution
    #     # y_diff = np.abs(null_grid['lat'].diff('lat', 1).mean().round(5).values)
    #     # x_diff = np.abs(null_grid['lon'].diff('lon', 1).mean().round(5).values)

    #     # grid_res = np.round(np.mean([y_diff, x_diff]), 5)

    #     pass


    def load_dataset_metadata(self, datasets):
        """

        """
        super().load_dataset_metadata(datasets)

        dataset_list = self.dataset_list

        ## Checks
        grid_bool = all([d['result_type'] == 'grid' for d in dataset_list])
        if not grid_bool:
            raise ValueError('All values of result_type in the datasets should be grid.')

        grouping_len = len(np.unique([d['grouping'] for d in dataset_list]))
        if grouping_len > 1:
            raise ValueError('All values of grouping in the datasets must be the same.')

        grouping = dataset_list[0]['grouping']

        # param_len = len(np.unique([d['parameter'] for d in dataset_list]))
        # if param_len > 1:
        #     raise ValueError('All values of parameter in the datasets must be the same.')

        setattr(self, 'grouping', grouping)

        return dataset_list


    def parse_paths(self, glob_obj, date_format=None, from_date=None, to_date=None):
        """

        """
        paths = parse_data_paths(glob_obj, date_format, from_date, to_date)

        self.source_data_paths = paths

        return paths


    def load_results(self, block_size, results=None, dataset_id=None, sum_closed='right', other_closed='left', discrete=True, other_attrs=None, other_encoding=None, run_date=None, skip_resampling=True):
        """

        """
        ## Open datasets
        if isinstance(results, list):
            # if isinstance(results[0], str):
            #     paths = []
            #     [paths.extend(glob(p)) for p in results]
            #     paths.sort()
            #     data_list = [xr.open_dataset(p) for p in paths if isinstance(p, str)]
            if isinstance(results[0], xr.Dataset):
                data_list = [xr.open_dataset(p) for p in results if isinstance(p, xr.Dataset)]
            else:
                raise TypeError('If results is a list, then it must be xr.Datasets.')
        # elif isinstance(results, str):
        #     paths = sorted(glob(results))
        #     data_list = [xr.open_dataset(p) for p in paths]
        elif isinstance(results, xr.Dataset):
            data_list = [results]
        else:
            try:
                data_list = [xr.open_dataset(p) for p in self.source_data_paths]
            except:
                raise TypeError('results must be a list of str/xr.Dataset, a xr.Dataset, or use the method parse_paths.')

        ## Determine dataset_ids
        if isinstance(dataset_id, str):
            data_dict = {dataset_id: data_list}
        else:
            if hasattr(self, 'source_data_paths'):
                ds_ids = set([os.path.split(p)[1].split('_part_')[0] for p in self.source_data_paths])
                data_dict = {ds: [] for ds in ds_ids}
                for i, p in enumerate(self.source_data_paths):
                    ds_id = os.path.split(p)[1].split('_part_')[0]
                    data_dict[ds_id].append(data_list[i])
            else:
                raise TypeError('If paths are not contained in results, then a dataset_id of a str must be passed.')

        ## Iterate through the datasets
        load_results = super().load_results

        for ds_id, data_list in data_dict.items():
            print(ds_id)

            n_files = len(data_list)

            for i, data in enumerate(data_list):
                print(str(i+1) + ' of ' + str(n_files) + ' chunks')
                data1 = data.copy()
                data1.load()
                data.close()
                data = None

                blocks = split_grid(data1, block_size, block_size)
                data1.close()
                data1 = None

                for block in blocks:
                    # Determine if all values are null
                    total_n_values = np.prod(list(block.dims.values()))
                    t1 = block.isnull().sum()
                    n_null = max([int(t1[t].values) for t in t1])

                    if n_null < total_n_values:
                        load_results(block, ds_id, sum_closed=sum_closed, other_closed=other_closed, discrete=discrete, other_attrs=other_attrs, other_encoding=other_encoding, run_date=run_date, skip_resampling=skip_resampling)


















